# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# classifiers.py
# Copyright (C) 2014-2020 Fracpete (pythonwekawrapper at gmail dot com)

import unittest
import weka.core.jvm as jvm
import weka.core.classes as classes
import weka.core.converters as converters
import weka.classifiers as classifiers
import weka.filters as filters
import wekatests.tests.weka_test as weka_test
import weka.attribute_selection as attsel


class TestClassifiers(weka_test.WekaTest):

    def test_instantiate_classifier(self):
        """
        Tests the instantiation of several classifier classes.
        """
        cname = "weka.classifiers.trees.J48"
        options = None
        cls = classifiers.Classifier(classname=cname, options=options)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname + "/" + str(options))
        self.assertEqual(cname, cls.classname, msg="Classnames differ!")

        cname = "weka.classifiers.trees.J48"
        options = ["-C", "0.3"]
        cls = classifiers.Classifier(classname=cname, options=options)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname + "/" + str(options))
        self.assertEqual(cname, cls.classname, msg="Classnames differ!")

        cname = "weka.classifiers.meta.FilteredClassifier"
        options = ["-W", "weka.classifiers.trees.J48", "--", "-C", "0.3"]
        cls = classifiers.SingleClassifierEnhancer(classname=cname, options=options)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname + "/" + str(options))
        self.assertEqual(cname, cls.classname, msg="Classnames differ!")
        fname = "weka.filters.unsupervised.attribute.Remove"
        flter = filters.Filter(classname=fname, options=["-R", "last"])
        cls.filter = flter
        self.assertEqual(fname, cls.filter.classname, msg="Classnames differ!")

        cls = classifiers.FilteredClassifier()
        self.assertIsNotNone(cls, msg="Failed to instantiate FilteredClassifier!")
        self.assertEqual("weka.classifiers.meta.FilteredClassifier", cls.classname, msg="Classnames differ!")

        cname = "weka.classifiers.functions.SMO"
        cls = classifiers.KernelClassifier(classname=cname)
        self.assertIsNotNone(cls, msg="Failed to instantiate KernelClassifier: " + cname)
        self.assertEqual(cname, cls.classname, msg="Classnames differ!")
        kname = "weka.classifiers.functions.supportVector.RBFKernel"
        kernel = classifiers.Kernel(classname=kname)
        self.assertIsNotNone(kernel, msg="Failed to instantiate Kernel: " + kname)
        cls.kernel = kernel
        self.assertEqual(kname, cls.kernel.classname, msg="Kernel classnames differ!")

        cname = "weka.classifiers.meta.Vote"
        cls = classifiers.MultipleClassifiersCombiner(classname=cname)
        self.assertIsNotNone(cls, msg="Failed to instantiate MultipleClassifiersCombiner: " + cname)
        self.assertEqual(cname, cls.classname, msg="Classnames differ!")

    def test_capabilities(self):
        """
        Tests the capabilities.
        """
        cname = "weka.classifiers.trees.J48"
        options = None
        cls = classifiers.Classifier(classname=cname, options=options)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname + "/" + str(options))
        self.assertEqual(cname, cls.classname, msg="Classnames differ!")

        caps = cls.capabilities
        self.assertIsNotNone(caps, msg="Capabilities are None!")

    def test_build_classifier(self):
        """
        Tests the build_classifier method.
        """
        # 1. nominal
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data)
        data.class_is_last()

        cname = "weka.classifiers.trees.J48"
        options = ["-C", "0.3"]
        cls = classifiers.Classifier(classname=cname, options=options)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname + "/" + str(options))

        cls.build_classifier(data)

        # 2. numeric
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("bolts.arff"))
        self.assertIsNotNone(data)
        data.class_is_last()

        cname = "weka.classifiers.functions.LinearRegression"
        options = ["-R", "0.1"]
        cls = classifiers.Classifier(classname=cname, options=options)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname + "/" + str(options))

        cls.build_classifier(data)

    def test_distribution_for_instance(self):
        """
        Tests the distribution_for_instance method.
        """
        # 1. nominal
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data)
        data.class_is_last()

        cname = "weka.classifiers.trees.J48"
        options = ["-C", "0.3"]
        cls = classifiers.Classifier(classname=cname, options=options)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname + "/" + str(options))

        cls.build_classifier(data)
        for i in range(10):
            dist = cls.distribution_for_instance(data.get_instance(i))
            self.assertIsNotNone(dist)
            self.assertEqual(6, len(dist), msg="Number of classes in prediction differ!")

        # 2. numeric
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("bolts.arff"))
        self.assertIsNotNone(data)
        data.class_is_last()

        cname = "weka.classifiers.functions.LinearRegression"
        options = ["-R", "0.1"]
        cls = classifiers.Classifier(classname=cname, options=options)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname + "/" + str(options))

        cls.build_classifier(data)
        for i in range(10):
            dist = cls.distribution_for_instance(data.get_instance(i))
            self.assertIsNotNone(dist)
            self.assertEqual(1, len(dist), msg="Number of classes in prediction should be one for numeric classifier!")

    def test_batchpredictor(self):
        """
        Tests the batch predictor methods.
        """

        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data)
        data.class_is_last()

        cname = "weka.classifiers.trees.J48"
        options = []
        cls = classifiers.Classifier(classname=cname, options=options)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname + "/" + str(options))

        # batch predictor?
        self.assertTrue(cls.is_batchpredictor, msg="not a batch predictor: " + cname + "/" + str(options))

        # more efficient implementation?
        cls.has_efficient_batch_prediction()

        # batch size
        self.assertIsNotNone(cls.batch_size, msg="batch size is not initialized")

        # distributions_for_instances
        cls.build_classifier(data)
        dists = cls.distributions_for_instances(data)
        self.assertIsNotNone(dists, msg="no distributions generated")
        self.assertEqual(len(dists), len(data), msg="number of predictions differ")
        self.assertEqual(len(dists[0]), data.class_attribute.num_values, msg="size of distribution array does not match number of classes")

    def test_classify_instance(self):
        """
        Tests the classify_instance method.
        """
        # 1. nominal
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data)
        data.class_is_last()

        cname = "weka.classifiers.trees.J48"
        options = ["-C", "0.3"]
        cls = classifiers.Classifier(classname=cname, options=options)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname + "/" + str(options))

        cls.build_classifier(data)
        preds = []
        for i in range(10,20):
            pred = cls.classify_instance(data.get_instance(i))
            self.assertIsNotNone(pred)
            preds.append(pred)
        self.assertEqual(
            [2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 5.0, 5.0, 2.0, 2.0],
            preds, msg="Classifications differ")

        # 2. numeric
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("bolts.arff"))
        self.assertIsNotNone(data)
        data.class_is_last()

        cname = "weka.classifiers.functions.LinearRegression"
        options = ["-R", "0.1"]
        cls = classifiers.Classifier(classname=cname, options=options)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname + "/" + str(options))

        cls.build_classifier(data)
        preds = []
        for i in range(10):
            pred = cls.classify_instance(data.get_instance(i))
            self.assertIsNotNone(pred)
            preds.append(pred)
        expected = [24.313, 33.359, 28.569, 26.365, 32.680, 29.149, 26.998, 22.971, 13.160, 7.394]
        for i in range(len(preds)):
            self.assertAlmostEqual(expected[i], preds[i], places=3, msg="Classifications differ")

    def test_costmatrix(self):
        """
        Tests the CostMatrix class.
        """
        cmatrix = classifiers.CostMatrix(num_classes=3)
        self.assertEqual(3, cmatrix.num_columns, msg="# of columns differ")
        self.assertEqual(3, cmatrix.num_rows, msg="# of rows differ")

        self.assertEqual(0, cmatrix.get_element(1, 1), msg="cell should be initialized with 0")
        cmatrix.set_element(1, 1, 0.1)
        self.assertEqual(0.1, cmatrix.get_element(1, 1), msg="cell value differs")

        matrixstr = "[1.0 2.0; 3.0 4.0]"
        cmatrix = classifiers.CostMatrix.parse_matlab(matrixstr)
        self.assertEqual(matrixstr, cmatrix.to_matlab(), msg="Matrix strings differ")

    def test_evaluation(self):
        """
        Tests the Evaluation class.
        """
        # 1. nominal
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data)
        data.class_is_last()

        cname = "weka.classifiers.trees.J48"
        cls = classifiers.Classifier(classname=cname)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname)

        evl = classifiers.Evaluation(data)
        self.assertIsNotNone(evl, msg="Failed to instantiate Evaluation")
        self.assertIsNone(evl.header.equal_headers(data))
        self.assertEqual([9, 100, 685, 1, 68, 41], evl.class_priors.tolist())

        # cross-validate
        evl.crossvalidate_model(cls, data, 10, classes.Random(1))
        self.assertEqual([[5.0, 0.0, 3.0, 0.0, 0.0, 0.0], [0.0, 99.0, 0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 680.0, 0.0, 0.0, 2.0], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0, 67.0, 0.0], [0.0, 0.0, 7.0, 0.0, 0.0, 33.0]], evl.confusion_matrix.tolist(), msg="confusion matrix differs")
        self.assertAlmostEqual(0.749, evl.area_under_prc(0), places=3, msg="area_under_prc(0) differs")
        self.assertAlmostEqual(0.931, evl.area_under_roc(0), places=3, msg="area_under_roc(0) differs")
        self.assertAlmostEqual(0, evl.avg_cost, places=3, msg="avg_cost differs")
        self.assertEqual(884, evl.correct, msg="correct differs")
        self.assertAlmostEqual(98.775, evl.coverage_of_test_cases_by_predicted_regions, places=3, msg="coverage_of_test_cases_by_predicted_regions differs")
        self.assertAlmostEqual(0.016, evl.error_rate, places=3, msg="error_rate differs")
        self.assertAlmostEqual(0.769, evl.f_measure(0), places=3, msg="f_measure(0) differs")
        self.assertAlmostEqual(0.375, evl.false_negative_rate(0), places=3, msg="false_negative_rate(0) differs")
        self.assertAlmostEqual(0.0, evl.false_positive_rate(0), places=3, msg="false_positive_rate(0) differs")
        self.assertEqual(14, evl.incorrect, msg="incorrect differs")
        self.assertAlmostEqual(0.961, evl.kappa, places=3, msg="kappa differs")
        self.assertAlmostEqual(1017.196, evl.kb_information, places=3, msg="kb_information differs")
        self.assertAlmostEqual(1.133, evl.kb_mean_information, places=3, msg="kb_mean_information differs")
        self.assertAlmostEqual(95.017, evl.kb_relative_information, places=3, msg="kb_relative_information differs")
        self.assertAlmostEqual(0.789, evl.matthews_correlation_coefficient(0), places=3, msg="matthews_correlation_coefficient(0) differs")
        self.assertAlmostEqual(0.006, evl.mean_absolute_error, places=3, msg="mean_absolute_error differs")
        self.assertAlmostEqual(0.134, evl.mean_prior_absolute_error, places=3, msg="mean_prior_absolute_error differs")
        self.assertAlmostEqual(3.0, evl.num_false_negatives(0), places=3, msg="num_false_negatives(0) differs")
        self.assertAlmostEqual(0.0, evl.num_false_positives(0), places=3, msg="num_false_positives(0) differs")
        self.assertAlmostEqual(898.0, evl.num_instances, places=3, msg="num_instances differs")
        self.assertAlmostEqual(890.0, evl.num_true_negatives(0), places=3, msg="num_true_negatives(0) differs")
        self.assertAlmostEqual(5.0, evl.num_true_positives(0), places=3, msg="num_true_positives(0) differs")
        self.assertAlmostEqual(98.44, evl.percent_correct, places=2, msg="percent_correct differs")
        self.assertAlmostEqual(1.56, evl.percent_incorrect, places=2, msg="percent_incorrect differs")
        self.assertAlmostEqual(0, evl.percent_unclassified, places=2, msg="percent_unclassified differs")
        self.assertAlmostEqual(1.0, evl.precision(0), places=3, msg="precision(0) differs")
        self.assertAlmostEqual(0.625, evl.recall(0), places=3, msg="recall(0) differs")
        self.assertAlmostEqual(4.187, evl.relative_absolute_error, places=3, msg="relative_absolute_error differs")
        self.assertAlmostEqual(0.258, evl.root_mean_prior_squared_error, places=3, msg="root_mean_prior_squared_error differs")
        self.assertAlmostEqual(0.067, evl.root_mean_squared_error, places=3, msg="root_mean_squared_error differs")
        self.assertAlmostEqual(25.912, evl.root_relative_squared_error, places=3, msg="root_relative_squared_error differs")
        self.assertAlmostEqual(-75.044, evl.sf_entropy_gain, places=3, msg="sf_entropy_gain differs")
        self.assertAlmostEqual(-0.084, evl.sf_mean_entropy_gain, places=3, msg="sf_mean_entropy_gain differs")
        self.assertAlmostEqual(1.192, evl.sf_mean_prior_entropy, places=3, msg="sf_mean_prior_entropy differs")
        self.assertAlmostEqual(1.276, evl.sf_mean_scheme_entropy, places=3, msg="sf_mean_scheme_entropy differs")
        self.assertAlmostEqual(16.722, evl.size_of_predicted_regions, places=3, msg="size_of_predicted_regions differs")
        self.assertEqual(0, evl.unclassified, msg="unclassified differs")
        self.assertAlmostEqual(0.0, evl.total_cost, places=3, msg="total_cost differs")
        self.assertAlmostEqual(0.926, evl.unweighted_macro_f_measure, places=3, msg="unweighted_macro_f_measure differs")
        self.assertAlmostEqual(0.984, evl.unweighted_micro_f_measure, places=3, msg="unweighted_micro_f_measure differs")
        self.assertAlmostEqual(0.992, evl.weighted_area_under_prc, places=3, msg="weighted_area_under_prc differs")
        self.assertAlmostEqual(0.995, evl.weighted_area_under_roc, places=3, msg="weighted_area_under_roc differs")
        self.assertAlmostEqual(0.984, evl.weighted_f_measure, places=3, msg="weighted_f_measure differs")
        self.assertAlmostEqual(0.016, evl.weighted_false_negative_rate, places=3, msg="weighted_false_negative_rate differs")
        self.assertAlmostEqual(0.036, evl.weighted_false_positive_rate, places=3, msg="weighted_false_positive_rate differs")
        self.assertAlmostEqual(0.959, evl.weighted_matthews_correlation, places=3, msg="weighted_matthews_correlation differs")
        self.assertAlmostEqual(0.984, evl.weighted_precision, places=3, msg="weighted_precision differs")
        self.assertAlmostEqual(0.984, evl.weighted_recall, places=3, msg="weighted_recall differs")
        self.assertAlmostEqual(0.964, evl.weighted_true_negative_rate, places=3, msg="weighted_true_negative_rate differs")
        self.assertAlmostEqual(0.984, evl.weighted_true_positive_rate, places=3, msg="weighted_true_positive_rate differs")
        self.assertIsNotNone(evl.predictions, msg="Should have had predictions")
        self.assertEqual(data.num_instances, len(evl.predictions), msg="Number of predictions differ from training set size")

        # train/test
        evl = classifiers.Evaluation(data)
        self.assertIsNotNone(evl, msg="Failed to instantiate Evaluation")
        cname = "weka.classifiers.trees.J48"
        cls = classifiers.Classifier(classname=cname)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname)
        cls.build_classifier(data)
        evl.test_model(cls, data)
        self.assertAlmostEqual(99.777, evl.percent_correct, places=3, msg="percent_correct differs")

        # predictions: no
        cname = "weka.classifiers.trees.J48"
        cls = classifiers.Classifier(classname=cname)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname)
        evl = classifiers.Evaluation(data)
        evl.discard_predictions = True
        self.assertIsNotNone(evl, msg="Failed to instantiate Evaluation")
        evl.crossvalidate_model(cls, data, 10, classes.Random(1))
        self.assertIsNone(evl.predictions, msg="Should have had no predictions")

        # output
        cname = "weka.classifiers.trees.J48"
        cls = classifiers.Classifier(classname=cname)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname)
        evl = classifiers.Evaluation(data)
        cname = "weka.classifiers.evaluation.output.prediction.PlainText"
        pout = classifiers.PredictionOutput(classname=cname)
        self.assertEqual(cname, pout.classname, msg="Output classnames differ!")
        self.assertIsNotNone(evl, msg="Failed to instantiate Evaluation")
        evl.crossvalidate_model(cls, data, 10, classes.Random(1), output=pout)
        self.assertGreater(len(str(pout)), 0, msg="Should have generated output")

        # 2. numeric
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("bolts.arff"))
        self.assertIsNotNone(data)
        data.class_is_last()

        cname = "weka.classifiers.functions.LinearRegression"
        cls = classifiers.Classifier(classname=cname)
        self.assertIsNotNone(cls, msg="Failed to instantiate: " + cname )

        evl = classifiers.Evaluation(data)
        self.assertIsNotNone(evl, msg="Failed to instantiate Evaluation")
        self.assertIsNone(evl.header.equal_headers(data))

        # cross-validate
        evl.crossvalidate_model(cls, data, 10, classes.Random(1))
        self.assertAlmostEqual(0.919, evl.correlation_coefficient, places=3, msg="correlation_coefficient differs")
        self.assertAlmostEqual(10.697, evl.error_rate, places=3, msg="error_rate differs")

    def test_gridsearch(self):
        """
        Tests the GridSearch class.
        """
        gs = classifiers.GridSearch()

        self.assertEqual({"property": "C", "expression": "pow(BASE,I)", "min": -3.0, "max": 3.0, "step": 1.0, "base": 10.0}, gs.x, msg="x of grid differs")
        x = gs.x
        x["min"] = -1.0
        x["max"] = 2.0
        gs.x = x
        self.assertEqual({"property": "C", "expression": "pow(BASE,I)", "min": -1.0, "max": 2.0, "step": 1.0, "base": 10.0}, gs.x, msg="x of grid differs")

        self.assertEqual({"property": "kernel.gamma", "expression": "pow(BASE,I)", "min": -3.0, "max": 3.0, "step": 1.0, "base": 10.0}, gs.y, msg="y of grid differs")
        y = gs.y
        y["min"] = -1.0
        y["max"] = 2.0
        gs.y = y
        self.assertEqual({"property": "kernel.gamma", "expression": "pow(BASE,I)", "min": -1.0, "max": 2.0, "step": 1.0, "base": 10.0}, gs.y, msg="y of grid differs")

        cls = classifiers.Classifier(classname="weka.classifiers.functions.LinearRegression")
        gs.classifier = cls
        gs.evaluation = gs.tags_evaluation.find("RMSE")
        self.assertEqual("RMSE", str(gs.evaluation), "evaluation differs: " + str(gs.evaluation))

        gs.evaluation = "ACC"
        self.assertEqual("ACC", str(gs.evaluation), "evaluation differs: " + str(gs.evaluation))
        cls = classifiers.Classifier(classname="weka.classifiers.trees.J48")
        gs.classifier = cls

    def test_multisearch(self):
        """
        Tests the MultiSearch class.
        NB: multisearch-weka-package must be installed (https://github.com/fracpete/multisearch-weka-package).
        """
        ms = classifiers.MultiSearch()

        mparam = classes.MathParameter()
        mparam.prop = "classifier.kernel.gamma"
        mparam.minimum = -3.0
        mparam.maximum = 3.0
        mparam.step = 1.0
        mparam.base = 10.0
        mparam.expression = "pow(BASE,I)"
        lparam = classes.ListParameter()
        lparam.prop = "classifier.C"
        lparam.values = ["-2.0", "-1.0", "0.0", "1.0", "2.0"]
        ms.parameters = [mparam, lparam]
        self.assertEqual(str([mparam, lparam]), str(ms.parameters), msg="parameters differ")

        cls = classifiers.Classifier(
            classname="weka.classifiers.functions.SMOreg",
            options=["-K", "weka.classifiers.functions.supportVector.RBFKernel"])
        ms.classifier = cls
        self.assertEqual(cls.to_commandline(), ms.classifier.to_commandline(), msg="classifiers differ")

        cls = classifiers.Classifier(classname="weka.classifiers.functions.LinearRegression")
        ms.classifier = cls
        ms.evaluation = ms.tags_evaluation.find("RMSE")
        self.assertEqual("RMSE", str(ms.evaluation), "evaluation differs: " + str(ms.evaluation))

        ms.evaluation = "ACC"
        self.assertEqual("ACC", str(ms.evaluation), "evaluation differs: " + str(ms.evaluation))
        cls = classifiers.Classifier(classname="weka.classifiers.trees.J48")
        ms.classifier = cls

    def test_attributeselectedclassifier(self):
        """
        Tests the AttributeSelectedClassifier class.
        """
        asc = classifiers.AttributeSelectedClassifier()
        asc.evaluator = attsel.ASEvaluation(classname="weka.attributeSelection.CfsSubsetEval", options=["-P", "1", "-E", "1"])
        asc.search = attsel.ASSearch(classname="weka.attributeSelection.GreedyStepwise", options=["-B", "-T", "-1.7E308", "-N", "-1", "-num-slots", "1"])
        self.assertEqual("weka.attributeSelection.CfsSubsetEval -P 1 -E 1", asc.evaluator.to_commandline(), msg="evaluator command-lines differs")
        self.assertEqual("weka.attributeSelection.GreedyStepwise -B -T -1.7E308 -N -1 -num-slots 1", asc.search.to_commandline(), msg="search command-lines differs")


def suite():
    """
    Returns the test suite.
    :return: the test suite
    :rtype: unittest.TestSuite
    """
    return unittest.TestLoader().loadTestsFromTestCase(TestClassifiers)


if __name__ == '__main__':
    jvm.start(packages=True)   # necessary for multisearch
    unittest.TextTestRunner().run(suite())
    jvm.stop()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# capabilities.py
# Copyright (C) 2014-2016 Fracpete (pythonwekawrapper at gmail dot com)

import javabridge
from weka.core.classes import JavaObject, Enum
from weka.core.dataset import Attribute, Instances


class Capability(Enum):
    """
    Wrapper for a Capability.
    """

    def __init__(self, jobject=None, member=None):
        """
        Initializes the wrapper with the specified Capability object.

        :param jobject: the Capability object to wrap
        :type jobject: JB_Object
        :param member: the capability string to parse
        :type member: str
        """
        super(Capability, self).__init__(\
            jobject=jobject, enum="weka.core.Capabilities$Capability", member=member)

    @property
    def is_attribute(self):
        """
        Returns whether this capability is an attribute.

        :return: whether it is an attribute
        :rtype: bool
        """
        return javabridge.call(self.jobject, "isAttribute", "()Z")

    @property
    def is_attribute_capability(self):
        """
        Returns whether this capability is an attribute capability.

        :return: whether it is an attribute capability
        :rtype: bool
        """
        return javabridge.call(self.jobject, "isAttributeCapability", "()Z")

    @property
    def is_class(self):
        """
        Returns whether this capability is a class.

        :return: whether it is a class
        :rtype: bool
        """
        return javabridge.call(self.jobject, "isClass", "()Z")

    @property
    def is_class_capability(self):
        """
        Returns whether this capability is a class capability.

        :return: whether it is a class capability
        :rtype: bool
        """
        return javabridge.call(self.jobject, "isClassCapability", "()Z")

    @property
    def is_other_capability(self):
        """
        Returns whether this capability is an other capability.

        :return: whether it is an other capability
        :rtype: bool
        """
        return javabridge.call(self.jobject, "isOtherCapability", "()Z")


class Capabilities(JavaObject):
    """
    Wrapper for Capabilities.
    """

    def __init__(self, jobject=None, owner=None):
        """
        Initializes the wrapper with the specified Capabilities object.

        :param jobject: the Capabilities object to wrap, instantiates a new one if None
        :type jobject: JB_Object
        :param owner: the weka.core.CapabilitiesHandler object to use as owner, can be None
        :type owner: JB_Object or JavaObject
        """
        if jobject is None:
            if isinstance(owner, JavaObject):
                owner = owner.jobject
            jobject = javabridge.make_instance(
                "weka/core/Capabilities", "(Lweka/core/CapabilitiesHandler;)V", owner)
        Capabilities.enforce_type(jobject, "weka.core.Capabilities")
        super(Capabilities, self).__init__(jobject)

    @property
    def owner(self):
        """
        Returns the owner of these capabilities, if any.

        :return: the owner, can be None
        :rtype: JavaObject
        """
        obj = javabridge.call(self.jobject, "getOwner", "()Lweka/core/CapabilitiesHandler;")
        if obj is None:
            return None
        else:
            return JavaObject(jobject=obj)

    @owner.setter
    def owner(self, obj):
        """
        Sets the new owner.

        :param obj: the new owner
        :type obj: JavaObject or JB_Object
        """
        if isinstance(obj, JavaObject):
            obj = obj.jobject
        javabridge.call(self.jobject, "setOwner", "(Lweka/core/CapabilitiesHandler;)V", obj)

    def capabilities(self):
        """
        Returns all the capabilities.

        :return: all capabilities
        :rtype: list
        """
        result = []
        iterator = javabridge.iterate_java(javabridge.call(self.jobject, "capabilities", "()Ljava/util/Iterator;"))
        for c in iterator:
            result.append(Capability(c))
        return result

    def attribute_capabilities(self):
        """
        Returns all the attribute capabilities.

        :return: attribute capabilities
        :rtype: Capabilities
        """
        return Capabilities(
            javabridge.call(self.jobject, "getAttributeCapabilities", "()Lweka/core/Capabilities;"))

    def class_capabilities(self):
        """
        Returns all the class capabilities.

        :return: class capabilities
        :rtype: Capabilities
        """
        return Capabilities(
            javabridge.call(self.jobject, "getClassCapabilities", "()Lweka/core/Capabilities;"))

    def other_capabilities(self):
        """
        Returns all other capabilities.

        :return: all other capabilities
        :rtype: Capabilities
        """
        return Capabilities(
            javabridge.call(self.jobject, "getOtherCapabilities", "()Lweka/core/Capabilities;"))

    def dependencies(self):
        """
        Returns all the dependencies.

        :return: the dependency list
        :rtype: list
        """
        result = []
        iterator = javabridge.iterate_java(javabridge.call(self.jobject, "dependencies", "()Ljava/util/Iterator;"))
        for c in iterator:
            result.append(Capability(c))
        return result

    def handles(self, capability):
        """
        Returns whether the specified capability is set.

        :param capability: the capability to check
        :type capability: Capability
        :return: whether the capability is set
        :rtype: bool
        """
        return javabridge.call(
            self.jobject, "handles", "(Lweka/core/Capabilities$Capability;)Z", capability.jobject)

    def enable_all(self):
        """
        enables all capabilities.
        """
        javabridge.call(self.jobject, "enableAll", "()V")

    def enable_all_attributes(self):
        """
        enables all attributes.
        """
        javabridge.call(self.jobject, "enableAllAttributes", "()V")

    def enable_all_classes(self):
        """
        enables all classes.
        """
        javabridge.call(self.jobject, "enableAllClasses", "()V")

    def enable(self, capability):
        """
        enables the specified capability.

        :param capability: the capability to enable
        :type capability: Capability
        """
        javabridge.call(self.jobject, "enable", "(Lweka/core/Capabilities$Capability;)V", capability.jobject)

    def enable_all_attribute_dependencies(self):
        """
        enables all attribute dependencies.
        """
        javabridge.call(self.jobject, "enableAllAttributeDependencies", "()V")

    def enable_all_class_dependencies(self):
        """
        enables all class dependencies.
        """
        javabridge.call(self.jobject, "enableAllClassDependencies", "()V")

    def enable_dependency(self, capability):
        """
        enables the dependency of the given capability enabling NOMINAL_ATTRIBUTES also enables
        BINARY_ATTRIBUTES, UNARY_ATTRIBUTES and EMPTY_NOMINAL_ATTRIBUTES.

        :param capability: the dependency to enable
        :type capability: Capability
        """
        javabridge.call(
            self.jobject, "enableDependency", "(Lweka/core/Capabilities$Capability;)V", capability.jobject)

    def disable_all(self):
        """
        Disables all capabilities.
        """
        javabridge.call(self.jobject, "disableAll", "()V")

    def disable_all_attributes(self):
        """
        Disables all attributes.
        """
        javabridge.call(self.jobject, "disableAllAttributes", "()V")

    def disable_all_classes(self):
        """
        Disables all classes.
        """
        javabridge.call(self.jobject, "disableAllClasses", "()V")

    def disable(self, capability):
        """
        Disables the specified capability.

        :param capability: the capability to disable
        :type capability: Capability
        """
        javabridge.call(self.jobject, "disable", "(Lweka/core/Capabilities$Capability;)V", capability.jobject)

    def disable_all_attribute_dependencies(self):
        """
        Disables all attribute dependencies.
        """
        javabridge.call(self.jobject, "disableAllAttributeDependencies", "()V")

    def disable_all_class_dependencies(self):
        """
        Disables all class dependencies.
        """
        javabridge.call(self.jobject, "disableAllClassDependencies", "()V")

    def disable_dependency(self, capability):
        """
        Disables the dependency of the given capability Disabling NOMINAL_ATTRIBUTES also disables
        BINARY_ATTRIBUTES, UNARY_ATTRIBUTES and EMPTY_NOMINAL_ATTRIBUTES.

        :param capability: the dependency to disable
        :type capability: Capability
        """
        javabridge.call(
            self.jobject, "disableDependency", "(Lweka/core/Capabilities$Capability;)V", capability.jobject)

    def has_dependencies(self):
        """
        Returns whether any dependencies are set.

        :return: whether any dependecies are set
        :rtype: bool
        """
        return javabridge.call(self.jobject, "hasDependencies", "()Z")

    def has_dependency(self, capability):
        """
        Returns whether the specified dependency is set.

        :param capability: the capability to check
        :type capability: Capability
        :return: whether the dependency is set
        :rtype: bool
        """
        return javabridge.call(
            self.jobject, "hasDependency", "(Lweka/core/Capabilities$Capability;)Z", capability.jobject)

    def supports(self, capabilities):
        """
        Returns true if the currently set capabilities support at least all of the capabiliites of the given
        Capabilities object (checks only the enum!)

        :param capabilities: the capabilities to check
        :type capabilities: Capabilities
        :return: whether the current capabilities support at least the specified ones
        :rtype: bool
        """
        return javabridge.call(self.jobject, "supports", "(Lweka/core/Capabilities;)Z", capabilities.jobject)

    def supports_maybe(self, capabilities):
        """
        Returns true if the currently set capabilities support (or have a dependency) at least all of the
        capabilities of the given Capabilities object (checks only the enum!)

        :param capabilities: the capabilities to check
        :type capabilities: Capabilities
        :return: whether the current capabilities (potentially) support the specified ones
        :rtype: bool
        """
        return javabridge.call(self.jobject, "supportsMaybe", "(Lweka/core/Capabilities;)Z", capabilities.jobject)

    @property
    def min_instances(self):
        """
        Returns the minimum number of instances that must be supported.

        :return: the minimum number
        :rtype: int
        """
        return javabridge.call(self.jobject, "getMinimumNumberInstances", "()I")

    @min_instances.setter
    def min_instances(self, minimum):
        """
        Sets the minimum number of instances that must be supported.

        :param minimum: the minimum number
        :type minimum: int
        """
        javabridge.call(self.jobject, "setMinimumNumberInstances", "(I)V", minimum)

    def test_attribute(self, att, is_class=None, fail=False):
        """
        Tests whether the attribute meets the conditions.

        :param att: the Attribute to test
        :type att: Attribute
        :param is_class: whether this attribute is the class attribute
        :type is_class: bool
        :param fail: whether to fail with an exception in case the test fails
        :type fail: bool
        :return: whether the attribute meets the conditions
        :rtype: bool
        """
        if fail:
            if is_class is None:
                return javabridge.call(
                    self.jobject, "test", "(Lweka/core/Attribute;)Z", att.jobject)
            else:
                return javabridge.call(
                    self.jobject, "test", "(Lweka/core/Attribute;Z)Z", att.jobject, is_class)
        else:
            if is_class is None:
                return javabridge.call(
                    self.jobject, "testWithFail", "(Lweka/core/Attribute;)Z", att.jobject)
            else:
                return javabridge.call(
                    self.jobject, "testWithFail", "(Lweka/core/Attribute;Z)Z", att.jobject, is_class)

    def test_instances(self, data, from_index=None, to_index=None, fail=False):
        """
        Tests whether the dataset meets the conditions.

        :param data: the Instances to test
        :type data: Instances
        :param from_index: the first attribute to include
        :type from_index: int
        :param to_index: the last attribute to include
        :type to_index: int
        :return: wether the dataset meets the requirements
        :rtype: bool
        """
        if fail:
            if (from_index is None) or (to_index is None):
                return javabridge.call(
                    self.jobject, "test", "(Lweka/core/Instances;)Z", data.jobject)
            else:
                return javabridge.call(
                    self.jobject, "test", "(Lweka/core/Instances;II)Z", data.jobject, from_index, to_index)
        else:
            if (from_index is None) or (to_index is None):
                return javabridge.call(
                    self.jobject, "testWithFail", "(Lweka/core/Instances;)Z", data.jobject)
            else:
                return javabridge.call(
                    self.jobject, "testWithFail", "(Lweka/core/Instances;II)Z", data.jobject, from_index, to_index)

    @classmethod
    def for_instances(cls, data, multi=None):
        """
        returns a Capabilities object specific for this data. The minimum number of instances is not set, the check
        for multi-instance data is optional.

        :param data: the data to generate the capabilities for
        :type data: Instances
        :param multi: whether to check the structure, too
        :type multi: bool
        :return: the generated capabilities
        :rtype: Capabilities
        """
        if multi is None:
            return Capabilities(javabridge.static_call(
                "weka/core/Capabilities", "forInstances",
                "(Lweka/core/Instances;)Lweka/core/Capabilities;", data.jobject))
        else:
            return Capabilities(javabridge.static_call(
                "weka/core/Capabilities", "forInstances",
                "(Lweka/core/Instances;Z)Lweka/core/Capabilities;", data.jobject, multi))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# capabilities.py
# Copyright (C) 2014-2016 Fracpete (pythonwekawrapper at gmail dot com)

import unittest
import weka.core.jvm as jvm
import weka.core.capabilities as capabilities
import weka.classifiers as classifiers
import wekatests.tests.weka_test as weka_test


class TestCapabilities(weka_test.WekaTest):

    def test_capability(self):
        """
        Tests the Capability class.
        """
        member = "NOMINAL_ATTRIBUTES"
        cap = capabilities.Capability(member=member)
        self.assertIsNotNone(cap, msg="Failed to instantiate capability: " + member)
        self.assertEqual(member, cap.name, "Names differ")
        self.assertTrue(cap.is_attribute)
        self.assertTrue(cap.is_attribute_capability)
        self.assertFalse(cap.is_class)
        self.assertFalse(cap.is_class_capability)
        self.assertFalse(cap.is_other_capability)

    def test_capabilities(self):
        """Tests the Capabilities class.
        """
        caps = capabilities.Capabilities()
        self.assertIsNotNone(caps, msg="Failed to instantiate empty capabilities")
        self.assertEqual(0, len(caps.capabilities()), msg="Should have no capabilities")

        cls = classifiers.Classifier(classname="weka.classifiers.trees.J48")
        caps = capabilities.Capabilities(owner=cls)
        self.assertIsNotNone(caps, msg="Failed to instantiate empty capabilities")
        self.assertEqual(0, len(caps.capabilities()), msg="Should have no capabilities")
        self.assertIsNotNone(caps.owner, msg="Should have an owner")
        caps.owner = None
        self.assertIsNone(caps.owner, msg="Should have no owner")
        caps.owner = cls
        self.assertIsNotNone(caps.owner, msg="Should have an owner")

        cap = capabilities.Capability(member="NUMERIC_CLASS")
        caps.enable(cap)
        self.assertTrue(caps.handles(cap), "Should have capability: " + str(cap))
        self.assertFalse(caps.has_dependency(cap), "Should have no dependency: " + str(cap))
        self.assertEqual(0, len(caps.dependencies()), msg="Should have no dependencies")
        caps.enable_dependency(cap)
        self.assertTrue(caps.has_dependency(cap), "Should have dependency: " + str(cap))
        self.assertEqual(1, len(caps.dependencies()), msg="Should have one dependency")

        caps.disable_all()
        self.assertEqual(0, len(caps.capabilities()), msg="Should have no capabilities")
        self.assertEqual(0, len(caps.dependencies()), msg="Should have no dependencies")

        caps.enable_all()
        self.assertGreater(len(caps.capabilities()), 0, msg="Should have capabilities")
        self.assertGreater(len(caps.dependencies()), 0, msg="Should have dependencies")

        caps.disable_all()
        caps.enable_all_attribute_dependencies()
        self.assertEqual(0, len(caps.capabilities()), msg="Should have no capabilities")
        self.assertGreater(len(caps.dependencies()), 0, msg="Should have dependencies")

        caps.disable_all()
        caps.enable_all_attributes()
        self.assertGreater(len(caps.capabilities()), 0, msg="Should have capabilities")
        self.assertEqual(0, len(caps.dependencies()), msg="Should have no dependencies")

        caps.disable_all()
        caps.enable_all_class_dependencies()
        self.assertEqual(0, len(caps.capabilities()), msg="Should have no capabilities")
        self.assertGreater(len(caps.dependencies()), 0, msg="Should have dependencies")

        caps.disable_all()
        caps.enable_all_classes()
        self.assertGreater(len(caps.capabilities()), 0, msg="Should have capabilities")
        self.assertEqual(0, len(caps.dependencies()), msg="Should have no dependencies")


def suite():
    """
    Returns the test suite.
    :return: the test suite
    :rtype: unittest.TestSuite
    """
    return unittest.TestLoader().loadTestsFromTestCase(TestCapabilities)


if __name__ == '__main__':
    jvm.start()
    unittest.TextTestRunner().run(suite())
    jvm.stop()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# attribute_selection.py
# Copyright (C) 2014-2016 Fracpete (pythonwekawrapper at gmail dot com)

import unittest
import weka.core.jvm as jvm
import weka.core.converters as converters
import weka.attribute_selection as attribute_selection
import wekatests.tests.weka_test as weka_test


class TestAttributeSelection(weka_test.WekaTest):

    def test_instantiations(self):
        """
        Tests instantiating classes.
        """
        cname = "weka.attributeSelection.BestFirst"
        options = ["-D", "1", "-N", "5"]
        search = attribute_selection.ASSearch(classname=cname, options=options)
        self.assertIsNotNone(search, msg="Search should not be None: " + cname + "/" + str(options))
        self.assertEqual(cname, search.classname, msg="Classnames differ!")

        cname = "weka.attributeSelection.CfsSubsetEval"
        options = ["-P", "1", "-E", "1"]
        evaluation = attribute_selection.ASEvaluation(classname=cname, options=options)
        self.assertIsNotNone(evaluation, msg="Evaluation should not be None: " + cname + "/" + str(options))
        self.assertEqual(cname, evaluation.classname, msg="Classnames differ!")

        attsel = attribute_selection.AttributeSelection()
        self.assertIsNotNone(search, msg="AttributeSelection should not be None!")

        attsel.search(search)
        attsel.evaluator(evaluation)

    def test_capabilities(self):
        """
        Tests the capabilities.
        """
        cname = "weka.attributeSelection.CfsSubsetEval"
        options = ["-P", "1", "-E", "1"]
        evaluation = attribute_selection.ASEvaluation(classname=cname, options=options)
        self.assertIsNotNone(evaluation, msg="Evaluation should not be None: " + cname + "/" + str(options))
        self.assertEqual(cname, evaluation.classname, msg="Classnames differ!")

        caps = evaluation.capabilities
        self.assertIsNotNone(caps, msg="Capabilities are None!")

    def test_attribute_selection(self):
        """
        Tests attribute selection.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data)
        data.class_is_last()

        cname = "weka.attributeSelection.BestFirst"
        options = ["-D", "1", "-N", "5"]
        search = attribute_selection.ASSearch(classname=cname, options=options)
        self.assertIsNotNone(search, msg="Search should not be None: " + cname + "/" + str(options))

        cname = "weka.attributeSelection.CfsSubsetEval"
        options = ["-P", "1", "-E", "1"]
        evaluation = attribute_selection.ASEvaluation(classname=cname, options=options)
        self.assertIsNotNone(evaluation, msg="Evaluation should not be None: " + cname + "/" + str(options))

        attsel = attribute_selection.AttributeSelection()
        self.assertIsNotNone(search, msg="AttributeSelection should not be None!")

        attsel.search(search)
        attsel.evaluator(evaluation)
        attsel.select_attributes(data)
        self.assertEqual(9, attsel.number_attributes_selected, msg="number_attributes_selected differs")
        self.assertEqual([0, 4, 8, 11, 12, 19, 24, 32, 33, 38], attsel.selected_attributes.tolist(), msg="selected_attributes differ")
        self.assertGreater(len(attsel.results_string), 0, msg="results_string should get produced")

    def test_reduce(self):
        """
        Tests reducing of attributes.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data)
        data.class_is_last()

        cname = "weka.attributeSelection.BestFirst"
        options = ["-D", "1", "-N", "5"]
        search = attribute_selection.ASSearch(classname=cname, options=options)
        self.assertIsNotNone(search, msg="Search should not be None: " + cname + "/" + str(options))

        cname = "weka.attributeSelection.CfsSubsetEval"
        options = ["-P", "1", "-E", "1"]
        evaluation = attribute_selection.ASEvaluation(classname=cname, options=options)
        self.assertIsNotNone(evaluation, msg="Evaluation should not be None: " + cname + "/" + str(options))

        attsel = attribute_selection.AttributeSelection()
        self.assertIsNotNone(search, msg="AttributeSelection should not be None!")

        attsel.search(search)
        attsel.evaluator(evaluation)
        attsel.select_attributes(data)
        reduced = attsel.reduce_dimensionality(data)
        self.assertEqual(attsel.number_attributes_selected + 1, reduced.num_attributes, msg="number of attributes differs")
        self.assertEqual(data.num_instances, reduced.num_instances, msg="number of rows differs")

    def test_ranking(self):
        """
        Tests ranking of attributes.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data)
        data.class_is_last()

        search = attribute_selection.ASSearch(classname="weka.attributeSelection.Ranker", options=["-N", "-1"])
        self.assertIsNotNone(search, msg="Search should not be None!")

        evaluation = attribute_selection.ASEvaluation("weka.attributeSelection.InfoGainAttributeEval")
        self.assertIsNotNone(evaluation, msg="Evaluation should not be None!")

        attsel = attribute_selection.AttributeSelection()
        self.assertIsNotNone(attsel, msg="AttributeSelection should not be None!")

        attsel.ranking(True)
        attsel.folds(2)
        attsel.crossvalidation(True)
        attsel.seed(42)
        attsel.search(search)
        attsel.evaluator(evaluation)
        attsel.select_attributes(data)

        self.assertGreater(len(str(attsel.ranked_attributes)), 0, msg="results_string should get produced")
        self.assertGreater(len(attsel.results_string), 0, msg="results_string should get produced")


def suite():
    """
    Returns the test suite.
    :return: the test suite
    :rtype: unittest.TestSuite
    """
    return unittest.TestLoader().loadTestsFromTestCase(TestAttributeSelection)


if __name__ == '__main__':
    jvm.start()
    unittest.TextTestRunner().run(suite())
    jvm.stop()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# filters.py
# Copyright (C) 2014-2020 Fracpete (pythonwekawrapper at gmail dot com)

import unittest
import weka.core.jvm as jvm
import weka.core.converters as converters
import weka.core.dataset as dataset
import weka.filters as filters
import wekatests.tests.weka_test as weka_test
import weka.attribute_selection as attsel


class TestFilters(weka_test.WekaTest):

    def test_instantiate_classifier(self):
        """
        Tests the instantiation of several classifier classes.
        """
        cname = "weka.filters.unsupervised.attribute.Remove"
        options = None
        flter = filters.Filter(classname=cname, options=["-R", "1,3"])
        self.assertIsNotNone(flter, msg="Failed to instantiate: " + cname + "/" + str(options))
        self.assertEqual(cname, flter.classname, msg="Classnames differ!")

    def test_make_copy(self):
        """
        Tests the make_copy class method.
        """
        cname = "weka.filters.unsupervised.attribute.Remove"
        options = None
        flter = filters.Filter(classname=cname, options=["-R", "1,3"])
        self.assertIsNotNone(flter, msg="Failed to instantiate: " + cname + "/" + str(options))
        self.assertEqual(cname, flter.classname, msg="Classnames differ!")

        flter2 = filters.Filter.make_copy(flter)
        self.assertIsNotNone(flter2, msg="Failed to instantiate: " + cname + "/" + str(options))
        self.assertEqual(cname, flter2.classname, msg="Classnames differ!")

    def test_capabilities(self):
        """
        Tests the capabilities.
        """
        cname = "weka.classifiers.trees.J48"
        options = None
        flter = filters.Filter(classname="weka.filters.unsupervised.attribute.Remove", options=["-R", "1,3"])
        self.assertIsNotNone(flter, msg="Failed to instantiate: " + cname + "/" + str(options))

        caps = flter.capabilities
        self.assertIsNotNone(caps, msg="Capabilities are None!")

    def test_batch_filtering(self):
        """
        Tests the Filter.filter method.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data)

        flter = filters.Filter(classname="weka.filters.unsupervised.attribute.Remove", options=["-R", "1,3"])
        flter.inputformat(data)
        filtered = flter.filter(data)
        self.assertEqual(data.num_attributes - 2, filtered.num_attributes, msg="Number of attributes differ")
        self.assertEqual(data.num_instances, filtered.num_instances, msg="Number of instances differ")

        # multple files
        data = loader.load_file(self.datafile("reutersTop10Randomized_1perc_shortened-train.arff"))
        self.assertIsNotNone(data)
        data2 = loader.load_file(self.datafile("reutersTop10Randomized_1perc_shortened-test.arff"))
        self.assertIsNotNone(data2)

        flter = filters.Filter(classname="weka.filters.unsupervised.attribute.StringToWordVector")
        flter.inputformat(data)
        filtered = flter.filter([data, data2])
        self.assertIsNone(filtered[0].equal_headers(filtered[1]), msg="Headers should be compatible")

    def test_incremental_filtering(self):
        """
        Tests the Filter.input/output methods.
        """
        loader = converters.Loader(classname="weka.core.converters.ArffLoader")
        data = loader.load_file(self.datafile("anneal.arff"))
        self.assertIsNotNone(data)

        flter = filters.Filter(classname="weka.filters.unsupervised.attribute.Remove", options=["-R", "1,3"])
        flter.inputformat(data)
        filtered = flter.outputformat()
        self.assertIsNotNone(filtered)
        self.assertTrue(isinstance(filtered, dataset.Instances), msg="Should be Instances object")

        for inst in data:
            flter.input(inst)
            finst = flter.output()
            filtered.add_instance(finst)

        self.assertEqual(data.num_attributes - 2, filtered.num_attributes, msg="Number of attributes differ")
        self.assertEqual(data.num_instances, filtered.num_instances, msg="Number of instances differ")

    def test_attributeselectionfilter(self):
        """
        Tests the AttributeSelection filter class.
        """
        fas = filters.AttributeSelection()
        fas.evaluator = attsel.ASEvaluation(classname="weka.attributeSelection.CfsSubsetEval", options=["-P", "1", "-E", "1"])
        fas.search = attsel.ASSearch(classname="weka.attributeSelection.GreedyStepwise", options=["-B", "-T", "-1.7E308", "-N", "-1", "-num-slots", "1"])
        self.assertEqual("weka.attributeSelection.CfsSubsetEval -P 1 -E 1", fas.evaluator.to_commandline(), msg="evaluator command-lines differs")
        self.assertEqual("weka.attributeSelection.GreedyStepwise -B -T -1.7E308 -N -1 -num-slots 1", fas.search.to_commandline(), msg="search command-lines differs")

def suite():
    """
    Returns the test suite.
    :return: the test suite
    :rtype: unittest.TestSuite
    """
    return unittest.TestLoader().loadTestsFromTestCase(TestFilters)


if __name__ == '__main__':
    jvm.start()
    unittest.TextTestRunner().run(suite())
    jvm.stop()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# conversion.py
# Copyright (C) 2015-2016 Fracpete (pythonwekawrapper at gmail dot com)


from weka.flow.base import Configurable
import weka.core.classes as classes


class Conversion(Configurable):
    """
    Ancestor for conversions used by the 'Convert' transformer.
    """

    def __init__(self, config=None):
        """
        Initializes the conversion.

        :param config: dictionary of options to use
        :type config: dict
        """
        super(Conversion, self).__init__(config=config)
        self._input = None
        self._output = None

    def __str__(self):
        """
        Returns a short representation of the actor's setup.

        :return: the setup
        :rtype: str
        """
        return classes.get_classname(self) + ": " + str(self._config)

    def check_input(self, obj):
        """
        Performs checks on the input object. Raises an exception if unsupported.

        :param obj: the object to check
        :type obj: object
        """
        pass

    @property
    def input(self):
        """
        Returns the current input object, None if not available.

        :return: the input object
        :rtype: object
        """
        return self._input

    @input.setter
    def input(self, obj):
        """
        Accepts the data for processing.

        :param obj: the object to process
        :type obj: object
        """
        self.check_input(obj)
        self._input = obj

    @property
    def output(self):
        """
        Returns the generated output object, None if not available.

        :return: the output object
        :rtype: object
        """
        return self._output

    def convert(self):
        """
        Performs the actual conversion.

        :return: None if successful, otherwise errors message
        :rtype: str
        """
        raise Exception("Not implemented!")


class PassThrough(Conversion):
    """
    Dummy conversion, just passes through the data.
    """

    def __init__(self, config=None):
        """
        Initializes the conversion.

        :param config: dictionary of options to use
        :type config: dict
        """
        super(PassThrough, self).__init__(config=config)

    def description(self):
        """
        Returns the description for the conversion.

        :return: the description
        :rtype: str
        """
        return "Dummy conversion, just passes through the data."

    def convert(self):
        """
        Performs the actual conversion.

        :return: None if successful, otherwise errors message
        :rtype: str
        """
        self._output = self._input
        return None


class AnyToCommandline(Conversion):
    """
    Generates a commandline string, e.g., from a classifier.
    """

    def __init__(self, config=None):
        """
        Initializes the conversion.

        :param config: dictionary of options to use
        :type config: dict
        """
        super(AnyToCommandline, self).__init__(config=config)

    def description(self):
        """
        Returns the description for the conversion.

        :return: the description
        :rtype: str
        """
        return "Generates a commandline string, e.g., from a classifier."

    def convert(self):
        """
        Performs the actual conversion.

        :return: None if successful, otherwise errors message
        :rtype: str
        """
        self._output = classes.to_commandline(self._input)
        return None


class CommandlineToAny(Conversion):
    """
    Generates an object from a commandline string, e.g., from a classifier setup.
    """

    def __init__(self, config=None):
        """
        Initializes the conversion.

        :param config: dictionary of options to use
        :type config: dict
        """
        super(CommandlineToAny, self).__init__(config=config)

    def description(self):
        """
        Returns the description for the conversion.

        :return: the description
        :rtype: str
        """
        return "Generates an object from a commandline string, e.g., from a classifier setup."

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        options = super(CommandlineToAny, self).fix_config(options)

        opt = "wrapper"
        if opt not in options:
            options[opt] = "weka.core.classes.OptionHandler"
        if opt not in self.help:
            self.help[opt] = "The name of the wrapper class to use (string)."

        return options

    def check_input(self, obj):
        """
        Performs checks on the input object. Raises an exception if unsupported.

        :param obj: the object to check
        :type obj: object
        """
        if isinstance(obj, str):
            return
        if isinstance(obj, unicode):
            return
        raise Exception("Unsupported class: " + self._input.__class__.__name__)

    def convert(self):
        """
        Performs the actual conversion.

        :return: None if successful, otherwise errors message
        :rtype: str
        """
        cname = str(self.config["wrapper"])
        self._output = classes.from_commandline(self._input, classname=cname)
        return None



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# stopwords.py
# Copyright (C) 2015-2016 Fracpete (pythonwekawrapper at gmail dot com)

import unittest
import os
import weka.core.jvm as jvm
from weka.core.stopwords import Stopwords
import wekatests.tests.weka_test as weka_test


class TestStopwords(weka_test.WekaTest):

    def test_instantiation(self):
        """
        Tests instantiating a stopwords handle.
        """
        stopwords = Stopwords()
        self.assertIsNotNone(stopwords, msg="Stopwords should not be None")
        stopwords = Stopwords(classname="weka.core.stopwords.Rainbow")
        self.assertIsNotNone(stopwords, msg="Stopwords should not be None")

    def test_is_stopword(self):
        """
        Tests calling the "is_stopword" method.
        """
        stopwords = Stopwords(classname="weka.core.stopwords.Rainbow")
        self.assertIsNotNone(stopwords, msg="Stopwords should not be None")
        word = "the"
        is_stopword = stopwords.is_stopword(word)
        self.assertEqual(True, is_stopword, msg="Should be stopword")
        word = "word"
        is_stopword = stopwords.is_stopword(word)
        self.assertEqual(False, is_stopword, msg="Should not be stopword")


def suite():
    """
    Returns the test suite.
    :return: the test suite
    :rtype: unittest.TestSuite
    """
    return unittest.TestLoader().loadTestsFromTestCase(TestStopwords)


if __name__ == '__main__':
    jvm.start()
    unittest.TextTestRunner().run(suite())
    jvm.stop()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# dataset.py
# Copyright (C) 2014-2016 Fracpete (pythonwekawrapper at gmail dot com)

import unittest
import weka.core.jvm as jvm
import weka.core.converters as converters
import weka.core.dataset as dataset
import weka.plot.dataset as plot
import wekatests.tests.weka_test as weka_test


class TestDataset(weka_test.WekaTest):

    def test_scatter_plot(self):
        """
        Tests the scatter_plot method.
        """
        loader = converters.Loader("weka.core.converters.ArffLoader")
        iris_data = loader.load_file(self.datafile("iris.arff"))
        iris_data.class_is_last()
        plot.scatter_plot(
            iris_data, iris_data.attribute_by_name("petalwidth").index,
            iris_data.attribute_by_name("petallength").index,
            percent=50,
            wait=False)

    def test_matrix_plot(self):
        """
        Tests the matrix_plot method.
        """
        loader = converters.Loader("weka.core.converters.ArffLoader")
        iris_data = loader.load_file(self.datafile("iris.arff"))
        iris_data.class_is_last()
        plot.matrix_plot(iris_data, percent=50, title="Matrix plot iris", wait=False)


def suite():
    """
    Returns the test suite.
    :return: the test suite
    :rtype: unittest.TestSuite
    """
    return unittest.TestLoader().loadTestsFromTestCase(TestDataset)


if __name__ == '__main__':
    jvm.start()
    unittest.TextTestRunner().run(suite())
    jvm.stop()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# timeseries.py
# Copyright (C) 2021 Fracpete (pythonwekawrapper at gmail dot com)

import javabridge
import logging
from weka.core.classes import JavaObject, OptionHandler, Date, Enum, new_instance
from weka.core.dataset import Instances, Instance
from weka.core.typeconv import jstring_list_to_string_list, jdouble_to_float
from weka.classifiers import Classifier, NumericPrediction
from weka.filters import Filter


# logging setup
logger = logging.getLogger("weka.timeseries")


class TestPart(JavaObject):
    """
    Inner class defining one boundary of an interval.
    """

    def __init__(self, jobject):
        """
        Initializes the TestPart object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        """
        self.enforce_type(jobject, "weka.classifiers.timeseries.core.CustomPeriodicTest.TestPart")
        super(TestPart, self).__init__(jobject=jobject)

    @property
    def is_upper(self):
        """
        Returns true if this is the upper bound.

        :return: true if upper bound
        :rtype: bool
        """
        return javabridge.call(self.jobject, "isUpper", "()Z")

    @is_upper.setter
    def is_upper(self, upper):
        """
        Sets whether this is the upper bound.

        :param upper: true if upper bound
        :type upper: bool
        """
        javabridge.call(self.jobject, "setIsUpper", "(Z)V")

    def eval(self, date, other):
        """
        Evaluate the supplied date against this bound. Handles
        date fields that are cyclic (such as month, day of week etc.)
        so that intervals such as oct < date < mar evaluate correctly.

        :param date: the date to test
        :type date: Date
        :param other: the other bound
        :type other: TestPart
        :return: true if the supplied date is within this bound
        :rtype: bool
        """
        return javabridge.call(self.jobject, "eval", "(Ljava/util/Date;Lweka/classifiers/timeseries/core/CustomPeriodicTest$TestPart;)Z", date.jobject, other.jobject)

    def operator(self, s):
        """
        Sets the operator.

        :param s: the operator to use
        :type s: str
        """
        javabridge.call(self.jobject, "setOperator", "(Ljava/lang/String;)V", s)

    def year(self, s):
        """
        Sets the year.

        :param s: the year to use
        :type s: str
        """
        javabridge.call(self.jobject, "setYear", "(Ljava/lang/String;)V", s)

    def week_of_year(self, s):
        """
        Sets the week of the year.

        :param s: the woy to use
        :type s: str
        """
        javabridge.call(self.jobject, "setWeekOfYear", "(Ljava/lang/String;)V", s)

    def week_of_month(self, s):
        """
        Sets the week of the month.

        :param s: the wom to use
        :type s: str
        """
        javabridge.call(self.jobject, "setWeekOfMonth", "(Ljava/lang/String;)V", s)

    def day_of_year(self, s):
        """
        Sets the day of year.

        :param s: the doy to use
        :type s: str
        """
        javabridge.call(self.jobject, "setDayOfYear", "(Ljava/lang/String;)V", s)

    def day_of_month(self, s):
        """
        Sets the day of the month.

        :param s: the dom to use
        :type s: str
        """
        javabridge.call(self.jobject, "setDayOfMonth", "(Ljava/lang/String;)V", s)

    @property
    def month(self):
        """
        Returns the month string.

        :return: the month string
        :rtype: str
        """
        return javabridge.call(self.jobject, "getMonthString", "()Ljava/lang/String;")

    @month.setter
    def month(self, s):
        """
        Sets the month.

        :param s: the month to use
        :type s: str
        """
        javabridge.call(self.jobject, "setMonth", "(Ljava/lang/String;)V", s)

    def day_of_week(self, s):
        """
        Sets the day of the week.

        :param s: the dow to use
        :type s: str
        """
        javabridge.call(self.jobject, "setDayOfWeek", "(Ljava/lang/String;)V", s)

    def hour_of_day(self, s):
        """
        Sets the hour of the day.

        :param s: the hod to use
        :type s: str
        """
        javabridge.call(self.jobject, "setHourOfDay", "(Ljava/lang/String;)V", s)

    def minute_of_hour(self, s):
        """
        Sets the minute of the hour.

        :param s: the moh to use
        :type s: str
        """
        javabridge.call(self.jobject, "setMinuteOfHour", "(Ljava/lang/String;)V", s)

    def second(self, s):
        """
        Sets the second.

        :param s: the second to use
        :type s: str
        """
        javabridge.call(self.jobject, "setSecond", "(Ljava/lang/String;)V", s)

    def day(self):
        """
        Returns the day string.

        :return: the day string
        :rtype: str
        """
        return javabridge.call(self.jobject, "getDayString", "()Ljava/lang/String;")


class CustomPeriodicTest(JavaObject):
    """
    Class that evaluates a supplied date against user-specified
    date constant fields. Fields that can be tested against include
    year, month, week of year, week of month, day of year, day of month,
    day of week, hour of day, minute of hour and second. Wildcard "*"
    matches any value for a particular field. Each CustomPeriodicTest
    is made up of one or two test parts. If the first test part's operator
    is "=", then no second part is necessary. Otherwise the first test part
    may use > or >= operators and the second test part < or <= operators.
    Taken together, the two parts define an interval. An optional label
    may be associated with the interval.
    """

    def __init__(self, jobject=None, test=None):
        """
        Initializes the CustomPeriodicTest object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param test: the test string to use
        :type test: str
        """
        if jobject is None:
            if test is None:
                raise Exception("Either jobject or test string must be provided!")
            else:
                jobject = javabridge.make_instance("weka.classifiers.timeseries.core.CustomPeriodicTest", "(Ljava/lang/String;)V", test)
        self.enforce_type(jobject, "weka.classifiers.timeseries.core.CustomPeriodicTest")
        super(CustomPeriodicTest, self).__init__(jobject=jobject)

    def lower_test(self):
        """
        Returns the lower bound test.

        :return: the test
        :rtype: TestPart
        """
        obj = javabridge.call(self.jobject, "getLowerTest", "()Lweka/classifiers/timeseries/core/CustomPeriodicTest$TestPart;")
        return TestPart(jobject=obj)

    def upper_test(self):
        """
        Returns the upper bound test.

        :return: the test
        :rtype: TestPart
        """
        obj = javabridge.call(self.jobject, "getUpperTest", "()Lweka/classifiers/timeseries/core/CustomPeriodicTest$TestPart;")
        return TestPart(jobject=obj)

    @property
    def label(self):
        """
        Returns the label.

        :return: the label
        :rtype: str
        """
        return javabridge.call(self.jobject, "getLabel", "()Ljava/lang/String;")

    @label.setter
    def label(self, l):
        """
        Sets the label.

        :param l: the label to use
        :type l: str
        """
        javabridge.call(self.jobject, "setLabel", "(Ljava/lang/String;)V", l)

    def test(self, test):
        """
        Sets the test as string.

        :param test: the test to use
        :type test: str
        """
        javabridge.call(self.jobject, "setTest", "(Ljava/lang/String;)V", test)

    def evaluate(self, date):
        """
        Evaluate the supplied date with respect to this custom periodic test interval.

        :param date: the date to test
        :type date: Date
        :return: true if the date lies within the interval.
        :rtype: bool
        """
        return javabridge.call(self.jobject, "evaluate", "(Ljava/util/Date;)Z", date.jobject)


class Periodicity(Enum):
    """
    Defines periodicity.
    """

    def __init__(self, jobject=None, periodicity=None):
        """
        Initializes the Periodicity object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param periodicity: the string representation of the enum
        :type periodicity: str
        """
        super(Periodicity, self).__init__(jobject=jobject, enum="weka.filters.supervised.attribute.TSLagMaker$Periodicity", member=periodicity)


class PeriodicityHandler(JavaObject):
    """
    Helper class to manage time stamp manipulation with respect to various
    periodicities. Has a routine to remap the time stamp, which is useful for
    date time stamps. Since dates are just manipulated internally as the number
    of milliseconds elapsed since the epoch, and any global trend modelling in
    regression functions results in enormous coefficients for this variable -
    remapping to a more reasonable scale prevents this. It also makes it easier
    to handle the case where there are time periods that shouldn't be
    considered as a time unit increment, e.g. weekends and public holidays for
    financial trading data. These "holes" in the data can be accomodated by
    accumulating a negative offset for the remapped date when a particular
    data/time occurs in a user-specified "skip" list.
    """

    def __init__(self, jobject):
        """
        Initializes the CustomPeriodicTest object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        """
        self.enforce_type(jobject, "weka.filters.supervised.attribute.TSLagMaker.PeriodicityHandler")
        super(PeriodicityHandler, self).__init__(jobject=jobject)

    @property
    def delta_time(self):
        """
        Returns the delta time.

        :return: the delta time
        :rtype: float
        """
        return javabridge.call(self.jobject, "deltaTime", "()D")

    @delta_time.setter
    def delta_time(self, value):
        """
        Sets the delta time.

        :param value: the delta time
        :type value: float
        """
        javabridge.call(self.jobject, "setDeltaTime", "(D)V")


class TSLagMaker(Filter):
    """
    A class for creating lagged versions of target variable(s) for use in time
    series forecasting. Uses the TimeseriesTranslate filter. Has options for
    creating averages of consecutive lagged variables (which can be useful for
    long lagged variables). Some polynomials of time are also created (if there
    is a time stamp), such as time^2 and time^3. Also creates cross products
    between time and the lagged and averaged lagged variables. If there is no
    date time stamp in the data then the user has the option of having an
    artificial time stamp created. Time stamps, real or otherwise, are used for
    modeling trends rather than using a differencing-based approach.

    Also has routines for dealing with a date timestamp - i.e. it can detect a
    monthly time period (because months are different lengths) and maps date time
    stamps to equal spaced time intervals. For example, in general, a date time
    stamp is remapped by subtracting the first observed value and adding this
    value divided by the constant delta (difference between consecutive steps) to
    the result. In the case of a detected monthly time period, the remapping
    involves subtracting the base year and then adding to this the number of the
    month within the current year plus twelve times the number of intervening
    years since the base year.

    Also has routines for adding new attributes derived from a date time stamp to
    the data - e.g. AM indicator, day of the week, month, quarter etc. In the
    case where there is no real date time stamp, the user may specify a nominal
    periodic variable (if one exists in the data). For example, month might be
    coded as a nominal value. In this case it can be specified as the primary
    periodic variable. The point is, that in all these cases (nominal periodic
    and date-derived periodics), we are able to determine what the value of these
    variables will be in future instances (as computed from the last known
    historic instance).
    """

    def __init__(self, jobject=None, options=None):
        """
        Initializes the TSLagMaker filter.

        :param jobject: the javaobject to use
        :type jobject: JB_Object
        :param options: the list of options to use
        :type options: list
        """
        super(TSLagMaker, self).__init__(jobject=jobject, classname="weka.filters.supervised.attribute.TSLagMaker", options=options)

    def clear_custom_periodics(self):
        """
        Clears the custom periodics.
        """
        javabridge.call(self.jobject, "clearCustomPeriodics", "()V")

    def add_custom_periodic(self, periodic):
        """
        Adds the custom periodic.

        :param periodic: the periodic to add
        :type periodic: str
        """
        javabridge.call(self.jobject, "addCustomPeriodic", "(Ljava/lang/String;)V", periodic)

    @property
    def fields_to_lag(self):
        """
        Returns the fields to lag as list.

        :return: the fields to lag
        :rtype: list
        """
        return jstring_list_to_string_list(javabridge.call(self.jobject, "getFieldsToLag", "()Ljava/util/List;"))

    @fields_to_lag.setter
    def fields_to_lag(self, fields):
        """
        Sets the fields to lag.

        :param fields: the list of fields to lag
        :type fields: list
        """
        javabridge.call(self.jobject, "setFieldsToLag", "(Ljava/util/List;)V", fields)

    @property
    def fields_to_lag_as_string(self):
        """
        Returns the fields to lag as string.

        :return: the fields to lag
        :rtype: str
        """
        return javabridge.call(self.jobject, "getFieldsToLagAsString", "()Ljava/lang/String;")

    @fields_to_lag_as_string.setter
    def fields_to_lag_as_string(self, fields):
        """
        Sets the fields to lag as string.

        :param fields: the fields to lag
        :type fields: str
        """
        javabridge.call(self.jobject, "setFieldsToLagAsString", "(Ljava/lang/String;)V", fields)

    @property
    def overlay_fields(self):
        """
        Returns the overlay fields as list.

        :return: the overlay fields
        :rtype: list
        """
        return jstring_list_to_string_list(javabridge.call(self.jobject, "getOverlayFields", "()Ljava/util/List;"))

    @overlay_fields.setter
    def overlay_fields(self, fields):
        """
        Sets the overlay fields.

        :param fields: the list of overlay fields
        :type fields: list
        """
        javabridge.call(self.jobject, "setOverlayFields", "(Ljava/util/List;)V", fields)

    @property
    def timestamp_field(self):
        """
        Returns the overlay fields as list.

        :return: the overlay fields
        :rtype: list
        """
        return javabridge.call(self.jobject, "getTimeStampField", "()Ljava/lang/String;")

    @timestamp_field.setter
    def timestamp_field(self, field):
        """
        Sets the timestamp field.

        :param field: the field with the timestamp
        :type field: str
        """
        javabridge.call(self.jobject, "setTimeStampField", "(Ljava/lang/String;)V", field)

    @property
    def remove_leading_instances_with_unknown_lag_values(self):
        """
        Returns whether to remove instances with unknown lag values.

        :return: true if to remove
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getRemoveLeadingInstancesWithUnknownLagValues", "()Z")

    @remove_leading_instances_with_unknown_lag_values.setter
    def remove_leading_instances_with_unknown_lag_values(self, remove):
        """
        Sets whether to remove instances with unknown lag values.

        :param remove: true if to remove
        :type remove: str
        """
        javabridge.call(self.jobject, "setRemoveLeadingInstancesWithUnknownLagValues", "(Z)V", remove)

    @property
    def adjust_for_trends(self):
        """
        Returns true if we are adjusting for trends via a real or artificial time stamp.

        :return: true if to adjust
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getAdjustForTrends", "()Z")

    @adjust_for_trends.setter
    def adjust_for_trends(self, adjust):
        """
        Sets whether we are adjusting for trends via a real or artificial time stamp.

        :param adjust: true if to adjust
        :type adjust: str
        """
        javabridge.call(self.jobject, "setAdjustForTrends", "(Z)V", adjust)

    @property
    def include_timelag_products(self):
        """
        Returns whether to include products between time and the lagged variables.

        :return: true if to include
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getIncludeTimeLagProducts", "()Z")

    @include_timelag_products.setter
    def include_timelag_products(self, include):
        """
        Sets whether to include products between time and the lagged variables.

        :param include: true if to include
        :type include: str
        """
        javabridge.call(self.jobject, "setIncludeTimeLagProducts", "(Z)V", include)

    @property
    def include_powers_of_time(self):
        """
        Returns whether to include powers of time in the transformed data.

        :return: true if to include
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getIncludePowersOfTime", "()Z")

    @include_powers_of_time.setter
    def include_powers_of_time(self, include):
        """
        Sets whether to include powers of time in the transformed data.

        :param include: true if to include
        :type include: str
        """
        javabridge.call(self.jobject, "setIncludePowersOfTime", "(Z)V", include)

    @property
    def adjust_for_variance(self):
        """
        Returns true if we are adjusting for variance by taking the log of the target(s).

        :return: true if to adjust
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getAdjustForVariance", "()Z")

    @adjust_for_variance.setter
    def adjust_for_variance(self, adjust):
        """
        Sets whether we are adjusting for variance by taking the log of the target(s).

        :param adjust: true if to adjust
        :type adjust: str
        """
        javabridge.call(self.jobject, "setAdjustForVariance", "(Z)V", adjust)

    @property
    def min_lag(self):
        """
        Returns the minimum lag to create.

        :return: the lag
        :rtype: int
        """
        return javabridge.call(self.jobject, "getMinLag", "()I")

    @min_lag.setter
    def min_lag(self, lag):
        """
        Sets the minimum lag to create.

        :param lag: the lag
        :type lag: int
        """
        javabridge.call(self.jobject, "setMinLag", "(I)V", lag)

    @property
    def max_lag(self):
        """
        Returns the maximum lag to create.

        :return: the lag
        :rtype: int
        """
        return javabridge.call(self.jobject, "getMaxLag", "()I")

    @max_lag.setter
    def max_lag(self, lag):
        """
        Sets the maximum lag to create.

        :param lag: the lag
        :type lag: int
        """
        javabridge.call(self.jobject, "setMaxLag", "(I)V", lag)

    @property
    def lag_range(self):
        """
        Returns the lag range to create.

        :return: the lag range
        :rtype: str
        """
        return javabridge.call(self.jobject, "getLagRange", "()Ljava/lang/String;")

    @lag_range.setter
    def lag_range(self, lag):
        """
        Sets the lag range to create.

        :param lag: the lag range
        :type lag: str
        """
        javabridge.call(self.jobject, "setLagRange", "(Ljava/lang/String;)V", lag)

    @property
    def average_consecutive_long_lags(self):
        """
        Returns true if consecutive long lagged variables are to be averaged.

        :return: true if to average
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getAverageConsecutiveLongLags", "()Z")

    @average_consecutive_long_lags.setter
    def average_consecutive_long_lags(self, average):
        """
        Sets whether to average consecutive long lagged variables. Setting this to
        true creates new variables that are averages of long lags and the original
        lagged variables involved are removed.

        :param average: true if to average
        :type average: str
        """
        javabridge.call(self.jobject, "setAverageConsecutiveLongLags", "(Z)V", average)

    @property
    def average_lags_after(self):
        """
        Returns the point after which long lagged variables will be averaged.

        :return: the lag
        :rtype: int
        """
        return javabridge.call(self.jobject, "getAverageLagsAfter", "()I")

    @average_lags_after.setter
    def average_lags_after(self, lag):
        """
        Sets the point after which long lagged variables will be averaged.

        :param lag: the lag
        :type lag: int
        """
        javabridge.call(self.jobject, "setAverageLagsAfter", "(I)V", lag)

    @property
    def num_consecutive_long_lags_to_average(self):
        """
        Returns the number of consecutive long lagged variables to average.

        :return: the lag
        :rtype: int
        """
        return javabridge.call(self.jobject, "getNumConsecutiveLongLagsToAverage", "()I")

    @num_consecutive_long_lags_to_average.setter
    def num_consecutive_long_lags_to_average(self, num):
        """
        Sets the number of consecutive long lagged variables to average.

        :param num: the lag
        :type num: int
        """
        javabridge.call(self.jobject, "setNumConsecutiveLongLagsToAverage", "(I)V", num)

    @property
    def primary_periodic_field_name(self):
        """
        Returns the name of the primary periodic attribute or null if one hasn't been specified.

        :return: the name
        :rtype: str
        """
        return javabridge.call(self.jobject, "getPrimaryPeriodicFieldName", "()Ljava/lang/String;")

    @primary_periodic_field_name.setter
    def primary_periodic_field_name(self, lag):
        """
        Sets the name of the primary periodic attribute or null if one hasn't been specified.

        :param lag: the name
        :type lag: str
        """
        javabridge.call(self.jobject, "setPrimaryPeriodicFieldName", "(Ljava/lang/String;)V", lag)

    @property
    def add_am_indicator(self):
        """
        Returns whether to add an AM indicator.

        :return: true if to add
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getAddAMIndicator", "()Z")

    @add_am_indicator.setter
    def add_am_indicator(self, add):
        """
        Sets whether to add an AM indicator.

        :param add: true if to add
        :type add: bool
        """
        javabridge.call(self.jobject, "setAddAMIndicator", "(Z)V", add)

    @property
    def add_day_of_week(self):
        """
        Returns whether to add day of week attribute.

        :return: true if to add
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getAddDayOfWeek", "()Z")

    @add_day_of_week.setter
    def add_day_of_week(self, add):
        """
        Sets whether to add day of week attribute.

        :param add: true if to add
        :type add: bool
        """
        javabridge.call(self.jobject, "setAddDayOfWeek", "(Z)V", add)

    @property
    def add_day_of_month(self):
        """
        Returns whether to add day of month attribute.

        :return: true if to add
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getAddDayOfMonth", "()Z")

    @add_day_of_month.setter
    def add_day_of_month(self, add):
        """
        Sets whether to add day of month attribute.

        :param add: true if to add
        :type add: bool
        """
        javabridge.call(self.jobject, "setAddDayOfMonth", "(Z)V", add)

    @property
    def add_num_days_in_month(self):
        """
        Returns whether to add # of days in month attribute.

        :return: true if to add
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getAddNumDaysInMonth", "()Z")

    @add_num_days_in_month.setter
    def add_num_days_in_month(self, add):
        """
        Sets whether to add # of days in month attribute.

        :param add: true if to add
        :type add: bool
        """
        javabridge.call(self.jobject, "setAddNumDaysInMonth", "(Z)V", add)

    @property
    def add_weekend_indicator(self):
        """
        Returns whether to add a weekend indicator.

        :return: true if to add
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getAddWeekendIndicator", "()Z")

    @add_weekend_indicator.setter
    def add_weekend_indicator(self, add):
        """
        Sets whether to add a weekend indicator.

        :param add: true if to add
        :type add: bool
        """
        javabridge.call(self.jobject, "setAddWeekendIndicator", "(Z)V", add)

    @property
    def add_month_of_year(self):
        """
        Returns whether to add month of year attribute.

        :return: true if to add
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getAddMonthOfYear", "()Z")

    @add_month_of_year.setter
    def add_month_of_year(self, add):
        """
        Sets whether to add month of year attribute.

        :param add: true if to add
        :type add: bool
        """
        javabridge.call(self.jobject, "setAddMonthOfYear", "(Z)V", add)

    @property
    def add_quarter_of_year(self):
        """
        Returns whether to add quarter of year attribute.

        :return: true if to add
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getAddQuarterOfYear", "()Z")

    @add_quarter_of_year.setter
    def add_quarter_of_year(self, add):
        """
        Sets whether to add quarter of year attribute.

        :param add: true if to add
        :type add: bool
        """
        javabridge.call(self.jobject, "setAddQuarterOfYear", "(Z)V", add)

    @property
    def is_using_artificial_time_index(self):
        """
        Returns whether an artifical time index is used.

        :return: true if to add
        :rtype: bool
        """
        return javabridge.call(self.jobject, "isUsingAnArtificialTimeIndex", "()Z")

    @property
    def artificial_time_start_value(self):
        """
        Returns the current value of the artificial time stamp. After training,
        after priming, and prior to forecasting, this will be equal to the number
        of training instances seen.

        :return: the start
        :rtype: float
        """
        return javabridge.call(self.jobject, "getArtificialTimeStartValue", "()D")

    @artificial_time_start_value.setter
    def artificial_time_start_value(self, start):
        """
        Sets the starting value for the artificial time stamp.

        :param start: the start
        :type start: float
        """
        javabridge.call(self.jobject, "setArtificialTimeStartValue", "(D)V", start)

    @property
    def current_timestamp_value(self):
        """
        Returns the current (i.e. most recent) time stamp value. Unlike an
        artificial time stamp, the value after training, after priming and before
        forecasting, will be equal to the time stamp of the most recent priming
        instance.

        :return: the timestamp value
        :rtype: float
        """
        return javabridge.call(self.jobject, "getCurrentTimeStampValue", "()D")

    def increment_artificial_time_value(self, increment):
        """
        Increment the artificial time value with the supplied increment value.

        :param increment: the increment
        :type increment: int
        """
        javabridge.call(self.jobject, "incrementArtificialTimeValue", "(I)V", increment)

    @property
    def delta_time(self):
        """
        Returns the difference between time values. This may be only approximate for
        periods based on dates. It is best to used date-based arithmetic in this
        case for incrementing/decrementing time stamps.

        :return: the delta
        :rtype: float
        """
        return javabridge.call(self.jobject, "getDeltaTime", "()D")

    @property
    def periodicity(self):
        """
        Returns the Periodicity representing the time stamp in use for this lag maker.
        If the lag maker is not adjusting for trends, or an artificial time stamp
        is being used, then null is returned.

        :return: the periodicity
        :rtype: Periodicity
        """
        return Periodicity(jobject=javabridge.call(self.jobject, "getPeriodicity", "()Lweka/filters/supervised/attribute/TSLagMaker$Periodicity;"))

    @periodicity.setter
    def periodicity(self, periodicity):
        """
        Sets the periodicity for the data. This is ignored if the lag maker is not
        adjusting for trends or is using an artificial time stamp. If not specified
        or set to Periodicity.UNKNOWN (the default) then heuristics will be used to
        try and automatically determine the periodicity.

        :param periodicity: the periodicity
        :type periodicity: Periodicity
        """
        javabridge.call(self.jobject, "setPeriodicity", "(Lweka/filters/supervised/attribute/TSLagMaker$Periodicity;)V", periodicity.jobject)

    @property
    def skip_entries(self):
        """
        Returns a list of time units to be 'skipped' - i.e. not considered as an
        increment. E.g financial markets don't trade on the weekend, so the
        difference between friday closing and the following monday closing is one
        time unit (and not three). Can accept strings such as "sat", "sunday",
        "jan", "august", or explicit dates (with optional formatting string) such
        as "2011-07-04@yyyy-MM-dd", or integers. Integers are interpreted with
        respect to the periodicity - e.g for daily data they are interpreted as day
        of the year; for hourly data, hour of the day; weekly data, week of the
        year.

        :return: the lag range
        :rtype: str
        """
        return javabridge.call(self.jobject, "getSkipEntries", "()Ljava/lang/String;")

    @skip_entries.setter
    def skip_entries(self, lag):
        """
        Sets the list of time units to be 'skipped' - i.e. not considered as an
        increment. E.g financial markets don't trade on the weekend, so the
        difference between friday closing and the following monday closing is one
        time unit (and not three). Can accept strings such as "sat", "sunday",
        "jan", "august", or explicit dates (with optional formatting string) such
        as "2011-07-04@yyyy-MM-dd", or integers. Integers are interpreted with
        respect to the periodicity - e.g for daily data they are interpreted as day
        of the year; for hourly data, hour of the day; weekly data, week of the
        year.

        :param lag: the lag range
        :type lag: str
        """
        javabridge.call(self.jobject, "setSkipEntries", "(Ljava/lang/String;)V", lag)

    def create_time_lag_cross_products(self, data):
        """
        Creates the cross-products.

        :param data: the data to create the cross-products for
        :type data: Instances
        :return: the cross-products
        :rtype: Instances
        """
        return Instances(javabridge.call(self.jobject, "createTimeLagCrossProducts", "(Lweka/core/Instances;)Lweka/core/Instances;", data.jobject))

    def transformed_data(self, data):
        """
        Returns the transformed data.

        :param data: the data to transform
        :type data: Instances
        :return: the transformed data
        :rtype: Instances
        """
        return Instances(javabridge.call(self.jobject, "getTransformedData", "(Lweka/core/Instances;)Lweka/core/Instances;", data.jobject))

    def clear_lag_histories(self):
        """
        Clears any history accumulated in the lag creating filters.
        """
        return javabridge.call(self.jobject, "clearLagHistories", "()V")


class TSForecaster(OptionHandler):
    """
    Wrapper class for timeseries forecasters.
    """

    def __init__(self, classname="weka.classifiers.timeseries.WekaForecaster", jobject=None, options=None):
        """
        Initializes the specified timeseries forecaster using either the classname or the supplied JB_Object.

        :param classname: the classname of the timeseries forecaster
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        if jobject is None:
            jobject = TSForecaster.new_instance(classname, options)
            if jobject is None:
                raise Exception(
                    "Failed to instantiate forecaster '%s' - is package 'timeseriesForecasting' installed and jvm started with package support?" % classname)
        self.enforce_type(jobject, "weka.classifiers.timeseries.TSForecaster")
        super(TSForecaster, self).__init__(jobject=jobject, options=options)

    @property
    def base_model_has_serializer(self):
        """
        Check whether the base learner requires special serialization.

        :return: True if base learner requires special serialization, false otherwise
        :rtype: bool
        """
        return javabridge.call(self.jobject, "baseModelHasSerializer", "()Z")

    def save_base_model(self, fname):
        """
        Saves the base model under the given filename.

        :param fname: the file to save the base model under
        :type fname: str
        """
        javabridge.call(self.jobject, "saveBaseModel", "(Ljava/lang/String;)V", fname)

    def load_base_model(self, fname):
        """
        Loads the base model from the given filename.

        :param fname: the file to load the base model from
        :type fname: str
        """
        javabridge.call(self.jobject, "loadBaseModel", "(Ljava/lang/String;)V", fname)

    @property
    def uses_state(self):
        """
        Check whether the base learner requires operations regarding state.

        :return: True if base learner uses state-based predictions, false otherwise
        :rtype: bool
        """
        return javabridge.call(self.jobject, "usesState", "()Z")

    def clear_previous_state(self):
        """
        Reset model state.
        """
        javabridge.call(self.jobject, "clearPreviousState", "()V")

    @property
    def previous_state(self):
        """
        Returns the previous state.

        :return: the state as list of JB_Object objects
        :rtype: list
        """
        return list(javabridge.get_collection_wrapper(javabridge.call(self.jobject, "getPreviousState", "()Ljava/util/List;")))

    @previous_state.setter
    def previous_state(self, state):
        """
        Sets the previous state.

        :param state: the state to set
        :type state: list
        """
        l = javabridge.JClassWrapper('java.util.ArrayList')()
        for obj in state:
            l.add(obj)
        javabridge.call(self.jobject, "setPreviousState", "(Ljava/util/List;)V")

    def serialize_state(self, fname):
        """
        Serializes the state under the given filename.

        :param fname: the file to serialize the state under
        :type fname: str
        """
        javabridge.call(self.jobject, "serializeState", "(Ljava/lang/String;)V", fname)

    def load_serialized_state(self, fname):
        """
        Loads the serialized state from the given filename.

        :param fname: the file to deserialize the state from
        :type fname: str
        """
        javabridge.call(self.jobject, "loadSerializedState", "(Ljava/lang/String;)V", fname)

    @property
    def algorithm_name(self):
        """
        Returns the name of the algorithm.

        :return: the name
        :rtype: str
        """
        return javabridge.call(self.jobject, "getAlgorithmName", "()Ljava/lang/String;")

    def reset(self):
        """
        Resets the algorithm.
        """
        javabridge.call(self.jobject, "reset", "()V")

    @property
    def fields_to_forecast(self):
        """
        Returns the fields to forecast.

        :return: the fields
        :rtype: str
        """
        return javabridge.call(self.jobject, "getFieldsToForecast", "()Ljava/lang/String;")

    @fields_to_forecast.setter
    def fields_to_forecast(self, fields):
        """
        Sets the fields to forecast.

        :param fields: the comma-separated string or list of fields to forecast
        :type fields: str or list
        """
        if isinstance(fields, list):
            fields = ",".join(fields)
        javabridge.call(self.jobject, "setFieldsToForecast", "(Ljava/lang/String;)V", fields)

    def build_forecaster(self, data):
        """
        Builds the forecaster using the provided data.

        :param data: the data to train with
        :type data: Instances
        """
        javabridge.call(self.jobject, "buildForecaster", "(Lweka/core/Instances;[Ljava/io/PrintStream;)V", data.jobject, [])

    def prime_forecaster(self, data):
        """
        Primes the forecaster using the provided data.

        :param data: the data to prime with
        :type data: Instances
        """
        javabridge.call(self.jobject, "primeForecaster", "(Lweka/core/Instances;)V", data.jobject)

    def forecast(self, steps):
        """
        Produce a forecast for the target field(s).
        Assumes that the model has been built and/or primed so that a forecast can be generated.

        :param steps: number of forecasted values to produce for each target. E.g. a value of 5 would produce a prediction for t+1, t+2, ..., t+5.
        :type steps: int
        :return: a List of Lists (one for each step) of forecasted values for each target (NumericPrediction objects)
        :rtype: list
        """
        objs1 = javabridge.get_collection_wrapper(javabridge.call(self.jobject, "forecast", "(I[Ljava/io/PrintStream;)Ljava/util/List;", steps, []))
        list1 = []
        for obj1 in objs1:
            list2 = []
            objs2 = javabridge.get_collection_wrapper(obj1)
            for obj2 in objs2:
                list2.append(NumericPrediction(obj2))
            list1.append(list2)
        return list1

    def run_forecaster(self, forecaster, options):
        """
        Builds the forecaster using the provided data.
        """
        javabridge.call(self.jobject, "runForecaster", "(Lweka/classifiers/timeseries/TSForecaster;[Ljava/lang/String;)V", forecaster.jobject, options)


class TSLagUser(JavaObject):
    """
    Wrapper class for TSLagUser objects.
    """

    def __init__(self, jobject):
        """
        Wraps the TSLagUser object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        """
        self.enforce_type(jobject, "weka.classifiers.timeseries.core.TSLagUser")
        super(TSLagUser, self).__init__(jobject=jobject)

    @property
    def tslag_maker(self):
        """
        Returns the base forecaster.

        :return: the base forecaster
        :rtype: Classifier
        """
        return TSLagMaker(jobject=javabridge.call(self.jobject, "getTSLagMaker", "()Lweka/filters/supervised/attribute/TSLagMaker;"))

    @tslag_maker.setter
    def tslag_maker(self, tslag_maker):
        """
        Sets the base forecaster.

        :param tslag_maker: the lag maker to use
        :type tslag_maker: TSLagMaker
        """
        javabridge.call(self.jobject, "setTSLagMaker", "(Lweka/filters/supervised/attribute/TSLagMaker;)V", tslag_maker.jobject)


class ConfidenceIntervalForecaster(JavaObject):
    """
    Wrapper class for ConfidenceIntervalForecaster objects.
    """

    def __init__(self, jobject):
        """
        Wraps a ConfidenceIntervalForecaster object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        """
        self.enforce_type(jobject, "weka.classifiers.timeseries.core.ConfidenceIntervalForecaster")
        super(ConfidenceIntervalForecaster, self).__init__(jobject=jobject)

    @property
    def calculate_conf_intervals_for_forecasts(self):
        """
        Returns the number of steps for which confidence intervals will be computed.

        :return: the steps
        :rtype: int
        """
        return javabridge.call(self.jobject, "getCalculateConfIntervalsForForecasts", "()I")

    @calculate_conf_intervals_for_forecasts.setter
    def calculate_conf_intervals_for_forecasts(self, steps):
        """
        Sets the number of steps for which confidence intervals will be computed.

        :param steps: the steps
        :type steps: int
        """
        javabridge.call(self.jobject, "setCalculateConfIntervalsForForecasts", "(I)V", steps)

    @property
    def is_producing_confidence_intervals(self):
        """
        Returns true if this forecaster is computing confidence limits for some or
        all of its future forecasts (i.e. getCalculateConfIntervalsForForecasts() >
        0).

        :return: true if confidence intervals are produced
        :rtype: bool
        """
        return javabridge.call(self.jobject, "isProducingConfidenceIntervals", "()I")

    @property
    def confidence_level(self):
        """
        Returns the confidence level in use for computing confidence intervals.

        :return: the level
        :rtype: float
        """
        return javabridge.call(self.jobject, "getOverlayFields", "()Ljava/lang/String;")

    @confidence_level.setter
    def confidence_level(self, level):
        """
        Sets the confidence level in use for computing confidence intervals.

        :param level: the level
        :type level: float
        """
        javabridge.call(self.jobject, "setOverlayFields", "(Ljava/lang/String;)V", level)


class OverlayForecaster(JavaObject):
    """
    Wrapper class for OverlayForecaster objects.
    """

    def __init__(self, jobject):
        """
        Wraps a OverlayForecaster object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        """
        self.enforce_type(jobject, "weka.classifiers.timeseries.core.OverlayForecaster")
        super(OverlayForecaster, self).__init__(jobject=jobject)

    @property
    def overlay_fields(self):
        """
        Returns the overlay fields as string.

        :return: the overlay fields
        :rtype: str
        """
        return javabridge.call(self.jobject, "getOverlayFields", "()Ljava/lang/String;")

    @overlay_fields.setter
    def overlay_fields(self, fields):
        """
        Sets the overlay fields as string.

        :param fields: the overlay fields
        :type fields: str
        """
        javabridge.call(self.jobject, "setOverlayFields", "(Ljava/lang/String;)V", fields)

    @property
    def is_using_overlay_data(self):
        """
        Returns true if overlay data has been used to train this forecaster, and
        thus is expected to be supplied for future time steps when making a
        forecast.

        :return:
        """
        return javabridge.call(self.jobject, "isUsingOverlayData", "()Z")

    def forecast_with_overlays(self, steps, overlays):
        """
        Produce a forecast for the target field(s). Assumes that the model has been built
        and/or primed so that a forecast can be generated. Also assumes that
        the forecaster has been told which attributes are to be considered
        "overlay" attributes in the data. Overlay data is data that the
        forecaster will be provided with when making a forecast into the future - i.e.
        it will be given the values of these attributes for future instances. The
        overlay data provided to this method should have the same structure as
        the original data used to train the forecaster - i.e. all original fields
        should be present, including the targets and time stamp field (if supplied).
        The values of targets will of course be missing ('?') since we want to forecast
        those. The time stamp values (if a time stamp is in use) may be provided, in which
        case the forecaster will use the time stamp values in the overlay instances. If
        the time stamp values are missing, then date arithmetic (for date time stamps) will
        be used to advance the time value beyond the last seen training value;
        similarly, for artificial time stamps or non-date time stamps, the computed
        time delta will be used to increment beyond the last seen training value.

        The number of instances in the overlay data should typically match the number
        of steps that have been requested for forecasting. If these differ, then
        overlay.numInstances() will be the number of steps forecasted.

        :param steps: number of forecasted values to produce for each target. E.g. a value of 5 would produce a prediction for t+1, t+2, ..., t+5.
        :type steps: int
        :param overlays: the overlay data to use
        :type overlays: Instances
        :return: a List of Lists (one for each step) of forecasted values for each target (NumericPrediction objects)
        :rtype: list
        """
        objs1 = javabridge.get_collection_wrapper(javabridge.call(self.jobject, "forecast", "(Lweka/core/Instances;I[Ljava/io/PrintStream;)Ljava/util/List;", steps, overlays.jobject, []))
        list1 = []
        for obj1 in objs1:
            list2 = []
            objs2 = javabridge.get_collection_wrapper(obj1)
            for obj2 in objs2:
                list2.append(NumericPrediction(obj2))
            list1.append(list2)
        return list1


class IncrementallyPrimeable(JavaObject):
    """
    Wrapper class for IncrementallyPrimeable objects.
    """

    def __init__(self, jobject):
        """
        Wraps a IncrementallyPrimeable object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        """
        self.enforce_type(jobject, "weka.classifiers.timeseries.core.IncrementallyPrimeable")
        super(IncrementallyPrimeable, self).__init__(jobject=jobject)

    def prime_forecaster_incremental(self, inst):
        """
        Primes the forecaster using the provided data.

        :param inst: the instance to prime with
        :type inst: Instance
        """
        javabridge.call(self.jobject, "primeForecasterIncremental", "(Lweka/core/Instance;)V", inst.jobject)


class WekaForecaster(TSForecaster, TSLagUser, ConfidenceIntervalForecaster, OverlayForecaster, IncrementallyPrimeable):
    """
    Wrapper class for Weka timeseries forecasters.
    """

    def __init__(self, jobject=None, options=None):
        """
        Initializes a Weka timeseries forecaster.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        if jobject is None:
            jobject = new_instance("weka.classifiers.timeseries.WekaForecaster")
        super(WekaForecaster, self).__init__(jobject=jobject, options=options)

    def clear_custom_periodics(self):
        """
        Clears the custom periodics.
        """
        javabridge.call(self.jobject, "clearCustomPeriodics", "()V")

    def add_custom_periodic(self, periodic):
        """
        Adds the custom periodic.

        :param periodic: the periodic to add
        :type periodic: str
        """
        javabridge.call(self.jobject, "addCustomPeriodic", "(Ljava/lang/String;)V", periodic)

    @property
    def base_forecaster(self):
        """
        Returns the base forecaster.

        :return: the base forecaster
        :rtype: Classifier
        """
        return Classifier(jobject=javabridge.call(self.jobject, "getBaseForecaster", "()Lweka/classifiers/Classifier;"))

    @base_forecaster.setter
    def base_forecaster(self, base_forecaster):
        """
        Sets the base forecaster.

        :param base_forecaster: the base forecaster to use
        :type base_forecaster: Classifier
        """
        javabridge.call(self.jobject, "setBaseForecaster", "(Lweka/classifiers/Classifier;)V", base_forecaster.jobject)


class TSEvalModule(JavaObject):
    """
    Wrapper for TSEvalModule objects.
    """

    def __init__(self, jobject):
        """
        Initializes the evaluation module.

        :param jobject: the object to initialize with
        :type jobject: JB_Object
        """
        super(TSEvalModule, self).__init__(jobject)

    def reset(self):
        """
        Resets the module.
        """
        javabridge.call(self.jobject, "reset", "()V")

    @property
    def eval_name(self):
        """
        Returns the name.
        """
        return javabridge.call(self.jobject, "getEvalName", "()Ljava/lang/String;")

    @property
    def description(self):
        """
        Returns the description.
        """
        return javabridge.call(self.jobject, "getDescription", "()Ljava/lang/String;")

    @property
    def definition(self):
        """
        Returns the description.
        """
        return javabridge.call(self.jobject, "getDefinition", "()Ljava/lang/String;")

    def evaluate_for_instance(self, pred, inst):
        """
        Evaluate the given forecast(s) with respect to the given test instance. Targets with missing values are ignored.

        :param pred: the numeric prediction
        :type pred: NumericPrediction
        :param inst: the instance
        :type inst: Instance
        """
        javabridge.call(self.jobject, "evaluateForInstance", "(Ljava/util/List;Lweka/core/Instance;)V", pred.jobject, inst.jobject)

    def calculate_measure(self):
        """
        Calculate the measure that this module represents.

        :return: the value of the measure for this module for each of the target(s).
        :rtype: ndarray
        """
        result = javabridge.call(self.jobject, "calculateMeasure", "()[D")
        return javabridge.get_env().get_double_array_elements(result)

    @property
    def summary(self):
        """
        Returns the description.
        """
        return javabridge.call(self.jobject, "toSummaryString", "()Ljava/lang/String;")

    @property
    def target_fields(self):
        """
        Returns the list of target fields.

        :return: the list of target fields
        :rtype: list
        """
        return jstring_list_to_string_list(javabridge.call(self.jobject, "getTargetFields", "()Ljava/util/List;"))

    @target_fields.setter
    def target_fields(self, fields):
        """
        Sets the list of target fields.

        :param fields: the list of target fields
        :type fields: list
        """
        javabridge.call(self.jobject, "setTargetFields", "(Ljava/util/List;)V", fields)

    def __str__(self):
        """
        Returns the name.
        """
        return self.eval_name

    @classmethod
    def module_list(cls):
        """
        Returns list of available modules.

        :return: the list of modules (TSEvalModule objects)
        :rtype: list
        """
        result = []
        objs = javabridge.get_collection_wrapper(javabridge.static_call(
            "Lweka/classifiers/timeseries/eval/TSEvalModuleHelper;", "getModuleList",
            "()Ljava/util/List;"))
        for obj in objs:
            result.append(TSEvalModule(obj))
        return result

    @classmethod
    def module(cls, name):
        """
        Returns the module with the specified name.

        :param name: the name of the module to return
        :type name: str
        :return: the TSEvalModule object
        :rtype: TSEvalModule
        """
        return TSEvalModule(javabridge.static_call(
            "Lweka/classifiers/timeseries/eval/TSEvalModuleHelper;", "getModule",
            "(Ljava/lang/String;)Ljava/lang/Object;", name))


class ErrorModule(TSEvalModule):
    """
    Wrapper for ErrorModule objects.
    """

    def __init__(self, jobject):
        """
        Initializes the error module.

        :param jobject: the object to initialize with
        :type jobject: JB_Object
        """
        super(ErrorModule, self).__init__(jobject)

    def counts_for_targets(self):
        """
        Returns the number of predicted, actual pairs for each target. Only
        entries that are non-missing for both actual and predicted contribute
        to the overall count.

        :return: the number of predicted, actual pairs for each target.
        :rtype: ndarray
        """
        result = javabridge.call(self.jobject, "countsForTargets", "()[D")
        return javabridge.get_env().get_double_array_elements(result)

    def errors_for_target(self, target):
        """
        Returns the list of the errors for the supplied target.

        :param target: the target to get the errors for
        :type target: str
        :return: the errors
        :rtype: list
        """
        result = []
        objs = javabridge.get_collection_wrapper(javabridge.call(self.jobject, "getErrorsForTarget", "(Ljava/lang/String;)Ljava/util/List;", target))
        for obj in objs:
            result.append(jdouble_to_float(obj))
        return result

    def predictions_for_target(self, target):
        """
        Returns the list of predictions for the target.

        :param target: the target to get the predictions for
        :type target: str
        :return: list of NumericPrediction
        :rtype: list
        """
        result = []
        objs = javabridge.get_collection_wrapper(javabridge.call(self.jobject, "getPredictionsForTarget", "(Ljava/lang/String;)Ljava/util/List;", target))
        for obj in objs:
            result.append(NumericPrediction(obj))
        return result

    def predictions_for_all_targets(self):
        """
        Returns the list of predictions for all targets.

        :return: list of list of NumericPrediction
        :rtype: list
        """
        result1 = []
        objs1 = javabridge.get_collection_wrapper(javabridge.call(self.jobject, "getPredictionsForAllTargets", "()Ljava/util/List;"))
        for obj1 in objs1:
            objs2 = javabridge.get_collection_wrapper(obj1)
            result2 = []
            for obj2 in objs2:
                result2.append(NumericPrediction(obj2))
            result1.append(result2)
        return result1


class TSEvaluation(JavaObject):
    """
    Evaluation class for timeseries forecasters.
    """

    def __init__(self, train, test_split_size=0.3, test=None):
        """
        Initializes a TSEvaluation object.

        :param train: the training data to use
        :type train: Instances
        :param test_split_size: the number or percentage of instances to hold out from the end of the training data to be test data.
        :type test_split_size: float
        :param test: the explicit test set to use, overrides the split size
        :type test: Instances
        """
        if test is None:
            jobject = javabridge.static_call(
                "Lweka/classifiers/timeseries/eval/TSEvaluationHelper;", "newInstance",
                "(Lweka/core/Instances;D)Ljava/lang/Object;",
                train.jobject, test_split_size)
        else:
            jobject = javabridge.static_call(
                "Lweka/classifiers/timeseries/eval/TSEvaluationHelper;", "newInstance",
                "(Lweka/core/Instances;Lweka/core/Instances;)Ljava/lang/Object;",
                train.jobject, test.jobject)
        super(TSEvaluation, self).__init__(jobject)

        # set up variables
        self._horizon = None
        self._rebuild_model_after_each_test_forecast_step = None

        # initialize with values from Java class
        self.horizon = 1
        self.rebuild_model_after_each_test_forecast_step = False

    @property
    def training_data(self):
        """
        Returns the training data.

        :return: the training data, None if none available
        :rtype: Instances
        """
        data = javabridge.call(self.jobject, "getTrainingData", "()Lweka/core/Instances;")
        if data is None:
            return None
        else:
            return Instances(data)

    @training_data.setter
    def training_data(self, data):
        """
        Sets the training data.

        :param data: the training data
        :type data: Instances
        """
        javabridge.call(self.jobject, "setTrainingData", "(Lweka/core/Instances;)V", data.jobject)

    @property
    def test_data(self):
        """
        Returns the test data.

        :return: the test data, None if none available
        :rtype: Instances
        """
        data = javabridge.call(self.jobject, "getTestData", "()Lweka/core/Instances;")
        if data is None:
            return None
        else:
            return Instances(data)

    @test_data.setter
    def test_data(self, data):
        """
        Sets the test data.

        :param data: the test data
        :type data: Instances
        """
        javabridge.call(self.jobject, "setTestData", "(Lweka/core/Instances;)V", data.jobject)

    @property
    def evaluate_on_training_data(self):
        """
        Returns whether to evaluate on the training data.

        :return: whether to evaluate
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getEvaluateOnTrainingData", "()Z")

    @evaluate_on_training_data.setter
    def evaluate_on_training_data(self, evaluate):
        """
        Sets whether whether to evaluate on to training data.

        :param evaluate: whether to evaluate
        :type evaluate: bool
        """
        javabridge.call(self.jobject, "setEvaluateOnTrainingData", "(Z)V", evaluate)

    @property
    def evaluate_on_test_data(self):
        """
        Returns whether to evaluate on the test data.

        :return: whether to evaluate
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getEvaluateOnTestData", "()Z")

    @evaluate_on_test_data.setter
    def evaluate_on_test_data(self, evaluate):
        """
        Sets whether whether to evaluate on to test data.

        :param evaluate: whether to evaluate
        :type evaluate: bool
        """
        javabridge.call(self.jobject, "setEvaluateOnTestData", "(Z)V", evaluate)

    @property
    def horizon(self):
        """
        Returns the number of steps to predict into the future.

        :return: the number of steps
        :rtype: int
        """
        return self._horizon

    @horizon.setter
    def horizon(self, steps):
        """
        Sets the number of steps to predict into the future.

        :param steps: the number of steps
        :type steps: int
        """
        javabridge.call(self.jobject, "setHorizon", "(I)V", steps)
        self._horizon = steps

    @property
    def prime_window_size(self):
        """
        Returns the size of the priming window, ie the number of historical instances to present before making a forecast.

        :return: the size
        :rtype: int
        """
        return javabridge.call(self.jobject, "getPrimeWindowSize", "()I")

    @prime_window_size.setter
    def prime_window_size(self, size):
        """
        Sets the size of the priming window, ie the number of historical instances to present before making a forecast.

        :param size: the size
        :type size: int
        """
        javabridge.call(self.jobject, "setPrimeWindowSize", "(I)V", size)

    @property
    def prime_for_test_data_with_test_data(self):
        """
        Returns whether evaluation for test data should begin by priming with the first
        x test data instances and then forecasting from step x + 1. This is the
        only option if there is no training data and a model has been deserialized
        from disk. If we have training data, and it occurs immediately before the
        test data in time, then we can prime with the last x instances from the
        training data.

        :return: whether to prime
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getPrimeForTestDataWithTestData", "()Z")

    @prime_for_test_data_with_test_data.setter
    def prime_for_test_data_with_test_data(self, prime):
        """
        Sets whether evaluation for test data should begin by priming with the first
        x test data instances and then forecasting from step x + 1. This is the
        only option if there is no training data and a model has been deserialized
        from disk. If we have training data, and it occurs immediately before the
        test data in time, then we can prime with the last x instances from the
        training data.

        :param prime: whether to prime
        :type prime: bool
        """
        javabridge.call(self.jobject, "setPrimeForTestDataWithTestData", "(Z)V", prime)

    @property
    def rebuild_model_after_each_test_forecast_step(self):
        """
        Returns whether the forecasting model should be rebuilt after each forecasting
        step on the test data using both the training data and test data up to the
        current instance.

        :return: whether to rebuild
        :rtype: bool
        """
        return self._rebuild_model_after_each_test_forecast_step

    @rebuild_model_after_each_test_forecast_step.setter
    def rebuild_model_after_each_test_forecast_step(self, rebuild):
        """
        Sets whether the forecasting model should be rebuilt after each forecasting
        step on the test data using both the training data and test data up to the
        current instance.

        :param rebuild: whether to rebuild
        :return: bool
        """
        javabridge.call(self.jobject, "setRebuildModelAfterEachTestForecastStep", "(Z)V", rebuild)
        self._rebuild_model_after_each_test_forecast_step = rebuild

    @property
    def forecast_future(self):
        """
        Returns whether we should generate a future forecast beyond the end of the
        training and/or test data.

        :return: whether to prime
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getForecastFuture", "()Z")

    @forecast_future.setter
    def forecast_future(self, prime):
        """
        Sets whether we should generate a future forecast beyond the end of the
        training and/or test data.

        :param prime: whether to prime
        :type prime: bool
        """
        javabridge.call(self.jobject, "setForecastFuture", "(Z)V", prime)

    @property
    def evaluation_modules(self):
        """
        Returns the list of evaluation modules in use.

        :return: list of TSEvalModule object
        :rtype: list
        """
        result = []
        objs = javabridge.get_collection_wrapper(javabridge.call(self.jobject, "getEvaluationModules", "()Ljava/util/List;"))
        for obj in objs:
            result.append(TSEvalModule(obj))
        return result

    @evaluation_modules.setter
    def evaluation_modules(self, modules):
        """
        Sets the evaluation modules to use (comma-separated list or list of TSEvalModule/str objects).

        :param modules: the evaluation modules (str or list)
        :type modules: str
        """
        if isinstance(modules, list):
            modules = [str(module) for module in modules]
            modules.remove("Error")  # requires manual removal
            modules = ",".join(modules)
        javabridge.call(self.jobject, "setEvaluationModules", "(Ljava/lang/String;)V", modules)

    def __str__(self):
        """
        Prints a simple summary of the evaluation setup.

        :return: the summary
        :rtype: str
        """
        eval_modules = [x.eval_name for x in self.evaluation_modules]

        result = "=== Evaluation setup ===\n\n"
        if self.training_data is not None:
            result += "Relation: " + self.training_data.relationname + "\n" \
                      + "# Training instances: " + str(self.training_data.num_instances) + "\n"
        if self.test_data is not None:
            result +="# Test instances: " + str(self.test_data.num_instances) + "\n"
        result += "Evaluate on training data: " + str(self.evaluate_on_training_data) + "\n" \
                  + "Evaluate on test data: " + str(self.evaluate_on_test_data) + "\n" \
                  + "Horizon: " + str(self.horizon) + "\n" \
                  + "Prime window size: " + str(self.prime_window_size) + "\n" \
                  + "Prime for test data with test data: " + str(self.prime_for_test_data_with_test_data) + "\n" \
                  + "Rebuild model after each test forecast step: " + str(self.rebuild_model_after_each_test_forecast_step) + "\n" \
                  + "Forecast future: " + str(self.forecast_future) + "\n" \
                  + "Evaluation modules: " + ", ".join(eval_modules) + "\n" \
                  + "\n"
        return result

    def evaluate(self, forecaster, build_model=True):
        """
        Evaluates the forecaster.

        :param forecaster: the forecaster to evaluate
        :type forecaster: TSForecaster
        :param build_model: whether to build the model as well
        :type build_model: bool
        """
        javabridge.call(
            self.jobject, "evaluateForecaster",
            "(Lweka/classifiers/timeseries/TSForecaster;Z[Ljava/io/PrintStream;)V",
            forecaster.jobject, build_model, [])

    def predictions_for_training_data(self, step_number):
        """
        Predictions for all targets for the specified step number on the training data.

        :param step_number: number of the step into the future to return predictions for
        :type step_number: int
        """
        return ErrorModule(javabridge.call(
            self.jobject, "getPredictionsForTrainingData",
            "(I)Lweka/classifiers/timeseries/eval/ErrorModule;",
            step_number))

    def predictions_for_test_data(self, step_number):
        """
        Predictions for all targets for the specified step number on the test data.

        :param step_number: number of the step into the future to return predictions for
        :type step_number: int
        """
        return ErrorModule(javabridge.call(
            self.jobject, "getPredictionsForTestData",
            "(I)Lweka/classifiers/timeseries/eval/ErrorModule;",
            step_number))

    def print_future_forecast_on_training_data(self, forecaster):
        """
        Print the forecasted values (for all targets) beyond the end of the training data.

        :param forecaster: the forecaster to use
        :type forecaster: TSForecaster
        :return: the forecasted values
        :rtype: str
        """
        return javabridge.call(
            self.jobject, "printFutureTrainingForecast",
            "(Lweka/classifiers/timeseries/TSForecaster;)Ljava/lang/String;", forecaster.jobject)

    def print_future_forecast_on_test_data(self, forecaster):
        """
        Print the forecasted values (for all targets) beyond the end of the test data.

        :param forecaster: the forecaster to use
        :type forecaster: TSForecaster
        :return: the forecasted values
        :rtype: str
        """
        return javabridge.call(
            self.jobject, "printFutureTestForecast",
            "(Lweka/classifiers/timeseries/TSForecaster;)Ljava/lang/String;", forecaster.jobject)

    def print_predictions_for_training_data(self, title, target_name, step_ahead, instance_number_offset=0):
        """
        Print the predictions for a given target at a given step-ahead level on the training data.

        :param title: the title for the output
        :type title: str
        :param target_name: the name of the target to print predictions for
        :type target_name: str
        :param step_ahead: the step-ahead level - e.g. 3 would print the 3-step-ahead predictions
        :type step_ahead: int
        :param instance_number_offset: the offset from the start of the training data from which to print actual and predicted values
        :type instance_number_offset: int
        :return: the predicted/actual values
        :rtype: str
        """
        return javabridge.call(
            self.jobject, "printPredictionsForTrainingData",
            "(Ljava/lang/String;Ljava/lang/String;II)Ljava/lang/String;", title, target_name, step_ahead, instance_number_offset)

    def print_predictions_for_test_data(self, title, target_name, step_ahead, instance_number_offset=0):
        """
        Print the predictions for a given target at a given step-ahead level on the test data.

        :param title: the title for the output
        :type title: str
        :param target_name: the name of the target to print predictions for
        :type target_name: str
        :param step_ahead: the step-ahead level - e.g. 3 would print the 3-step-ahead predictions
        :type step_ahead: int
        :param instance_number_offset: the offset from the start of the test data from which to print actual and predicted values
        :type instance_number_offset: int
        :return: the predicted/actual values
        :rtype: str
        """
        return javabridge.call(
            self.jobject, "printPredictionsForTestData",
            "(Ljava/lang/String;Ljava/lang/String;II)Ljava/lang/String;", title, target_name, step_ahead, instance_number_offset)

    def summary(self):
        """
        Generates a summary.

        :return: the summary
        :rtype: str
        """
        return javabridge.call(
            self.jobject, "toSummaryString", "()Ljava/lang/String;")

    @classmethod
    def evaluate_forecaster(cls, forecaster, args):
        """
        Evaluates the forecaster with the given options.

        :param forecaster: the forecaster instance to use
        :type forecaster: TSForecaster
        :param args: the command-line arguments to use
        :type args: list
        """
        javabridge.static_call(
            "Lweka/classifiers/timeseries/eval/TSEvaluation;", "evaluateForecaster",
            "(Lweka/classifiers/timeseries/TSForecaster;[Ljava/lang/String;)Ljava/lang/String;",
            forecaster.jobject, args)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# experiments.py
# Copyright (C) 2015-2016 Fracpete (pythonwekawrapper at gmail dot com)

import logging
import math
import weka.plot as plot
if plot.matplotlib_available:
    import matplotlib.pyplot as plt
from weka.experiments import ResultMatrix

# logging setup
logger = logging.getLogger(__name__)


def plot_experiment(mat, title="Experiment", axes_swapped=False, measure="Statistic", show_stdev=False,
                    key_loc="lower right", outfile=None, wait=True):
    """
    Plots the results from an experiment.

    :param mat: the result matrix to plot
    :type mat: ResultMatrix
    :param title: the title for the experiment
    :type title: str
    :param axes_swapped: whether the axes whether swapped ("sets x cls" or "cls x sets")
    :type axes_swapped: bool
    :param measure: the measure that is being displayed
    :type measure: str
    :param show_stdev: whether to show the standard deviation as error bar
    :type show_stdev: bool
    :param key_loc: the location string for the key
    :type key_loc: str
    :param outfile: the output file, ignored if None
    :type outfile: str
    :param wait: whether to wait for the user to close the plot
    :type wait: bool
    """

    if not plot.matplotlib_available:
        logger.error("Matplotlib is not installed, plotting unavailable!")
        return

    if not isinstance(mat, ResultMatrix):
        logger.error("Need to supply a result matrix!")
        return

    fig, ax = plt.subplots()
    if axes_swapped:
        ax.set_xlabel(measure)
        ax.set_ylabel("Classifiers")
    else:
        ax.set_xlabel("Classifiers")
        ax.set_ylabel(measure)
    ax.set_title(title)
    fig.canvas.set_window_title(title)
    ax.grid(True)
    ticksx = []
    ticks = []
    inc = 1.0 / float(mat.columns)
    for i in range(mat.columns):
        ticksx.append((i + 0.5) * inc)
        ticks.append("[" + str(i+1) + "]")
    plt.xticks(ticksx, ticks)
    plt.xlim([0.0, 1.0])
    for r in range(mat.rows):
        x = []
        means = []
        stdevs = []
        for c in range(mat.columns):
            mean = mat.get_mean(c, r)
            stdev = mat.get_stdev(c, r)
            if not math.isnan(mean):
                x.append((c + 0.5) * inc)
                means.append(mean)
                if not math.isnan(stdev):
                    stdevs.append(stdev)
        plot_label = mat.get_row_name(r)
        if show_stdev:
            ax.errorbar(x, means, yerr=stdevs, fmt='-o', ls="-", label=plot_label)
        else:
            ax.plot(x, means, "o-", label=plot_label)
    plt.draw()
    plt.legend(loc=key_loc, shadow=True)
    if outfile is not None:
        plt.savefig(outfile)
    if wait:
        plt.show()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# classifiers.py
# Copyright (C) 2014-2019 Fracpete (pythonwekawrapper at gmail dot com)

import javabridge
import logging
import weka.plot as plot
if plot.matplotlib_available:
    import matplotlib.pyplot as plt
from weka.core.classes import JavaObject, join_options
from weka.core.dataset import Instances
from weka.classifiers import Classifier, Evaluation, NumericPrediction, NominalPrediction

# logging setup
logger = logging.getLogger(__name__)


def plot_classifier_errors(predictions, absolute=True, max_relative_size=50, absolute_size=50, title=None,
                           outfile=None, wait=True, key_loc="lower center"):
    """
    Plots the classifers for the given list of predictions.

    TODO: click events http://matplotlib.org/examples/event_handling/data_browser.html

    :param predictions: the predictions to plot, use a dict to plot predictions of multiple classifiers
                        (keys are used as prefixes for plots)
    :type predictions: list or dict
    :param absolute: whether to use absolute errors as size or relative ones
    :type absolute: bool
    :param max_relative_size: the maximum size in point in case of relative mode
    :type max_relative_size: int
    :param absolute_size: the size in point in case of absolute mode
    :type absolute_size: int
    :param title: an optional title
    :type title: str
    :param outfile: the output file, ignored if None
    :type outfile: str
    :param wait: whether to wait for the user to close the plot
    :type wait: bool
    :param key_loc: the location string for the key
    :type key_loc: str
    """
    if not plot.matplotlib_available:
        logger.error("Matplotlib is not installed, plotting unavailable!")
        return

    if isinstance(predictions, list):
        multiple = {"": predictions}
    else:
        multiple = predictions

    actual = []
    predicted = []
    error = None
    cls = None
    ax = None
    for prefix in multiple:
        predictions = multiple[prefix]
        if prefix == "":
            prefix = None
        for pred in predictions:
            actual.append(pred.actual)
            predicted.append(pred.predicted)
            if isinstance(pred, NumericPrediction):
                if error is None:
                    error = []
                error.append(abs(pred.error))
            elif isinstance(pred, NominalPrediction):
                if cls is None:
                    cls = []
                if pred.actual != pred.predicted:
                    cls.append(1)
                else:
                    cls.append(0)
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_xlabel("actual")
            ax.set_ylabel("predicted")
            if title is None:
                title = "Classifier errors"
            ax.set_title(title)
            ax.plot(ax.get_xlim(), ax.get_ylim(), ls="--", c="0.3")
            ax.grid(True)
            fig.canvas.set_window_title(title)
        if error is None and cls is None:
            ax.scatter(actual, predicted, s=absolute_size, alpha=0.5, label=prefix)
        elif cls is not None:
            ax.scatter(actual, predicted, c=cls, s=absolute_size, alpha=0.5, label=prefix)
        elif error is not None:
            if not absolute:
                min_err = min(error)
                max_err = max(error)
                factor = (max_err - min_err) / max_relative_size
                for i in range(len(error)):
                    error[i] = error[i] / factor * max_relative_size
            ax.scatter(actual, predicted, s=error, alpha=0.5, label=prefix)
    plt.draw()
    if len(multiple) > 1:
        plt.legend(loc=key_loc, shadow=True)
    if outfile is not None:
        plt.savefig(outfile)
    if wait:
        plt.show()


def generate_thresholdcurve_data(evaluation, class_index):
    """
    Generates the threshold curve data from the evaluation object's predictions.

    :param evaluation: the evaluation to obtain the predictions from
    :type evaluation: Evaluation
    :param class_index: the 0-based index of the class-label to create the plot for
    :type class_index: int
    :return: the generated threshold curve data
    :rtype: Instances
    """
    jtc = JavaObject.new_instance("weka.classifiers.evaluation.ThresholdCurve")
    pred = javabridge.call(evaluation.jobject, "predictions", "()Ljava/util/ArrayList;")
    result = Instances(
        javabridge.call(jtc, "getCurve", "(Ljava/util/ArrayList;I)Lweka/core/Instances;", pred, class_index))
    return result


def get_thresholdcurve_data(data, xname, yname):
    """
    Retrieves x and y columns from  of the data generated by the weka.classifiers.evaluation.ThresholdCurve
    class.

    :param data: the threshold curve data
    :type data: Instances
    :param xname: the name of the X column
    :type xname: str
    :param yname: the name of the Y column
    :type yname: str
    :return: tuple of x and y arrays
    :rtype: tuple
    """
    xi = data.attribute_by_name(xname).index
    yi = data.attribute_by_name(yname).index
    x = []
    y = []
    for i in range(data.num_instances):
        inst = data.get_instance(i)
        x.append(inst.get_value(xi))
        y.append(inst.get_value(yi))
    return x, y


def get_auc(data):
    """
    Calculates the area under the ROC curve (AUC).

    :param data: the threshold curve data
    :type data: Instances
    :return: the area
    :rtype: float
    """
    return javabridge.static_call(
        "weka/classifiers/evaluation/ThresholdCurve", "getROCArea", "(Lweka/core/Instances;)D", data.jobject)


def get_prc(data):
    """
    Calculates the area under the precision recall curve (PRC).

    :param data: the threshold curve data
    :type data: Instances
    :return: the area
    :rtype: float
    """
    return javabridge.static_call(
        "weka/classifiers/evaluation/ThresholdCurve", "getPRCArea", "(Lweka/core/Instances;)D", data.jobject)


def plot_roc(evaluation, class_index=None, title=None, key_loc="lower right", outfile=None, wait=True):
    """
    Plots the ROC (receiver operator characteristics) curve for the given predictions.

    TODO: click events http://matplotlib.org/examples/event_handling/data_browser.html

    :param evaluation: the evaluation to obtain the predictions from
    :type evaluation: Evaluation
    :param class_index: the list of 0-based indices of the class-labels to create the plot for
    :type class_index: list
    :param title: an optional title
    :type title: str
    :param key_loc: the position string for the key
    :type key_loc: str
    :param outfile: the output file, ignored if None
    :type outfile: str
    :param wait: whether to wait for the user to close the plot
    :type wait: bool
    """
    if not plot.matplotlib_available:
        logger.error("Matplotlib is not installed, plotting unavailable!")
        return
    if class_index is None:
        class_index = [0]
    ax = None
    for cindex in class_index:
        data = generate_thresholdcurve_data(evaluation, cindex)
        head = evaluation.header
        area = get_auc(data)
        x, y = get_thresholdcurve_data(data, "False Positive Rate", "True Positive Rate")
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_xlabel("False Positive Rate")
            ax.set_ylabel("True Positive Rate")
            if title is None:
                title = "ROC"
            ax.set_title(title)
            ax.grid(True)
            fig.canvas.set_window_title(title)
            plt.xlim([-0.05, 1.05])
            plt.ylim([-0.05, 1.05])
        plot_label = head.class_attribute.value(cindex) + " (AUC: %0.4f)" % area
        ax.plot(x, y, label=plot_label)
        ax.plot(ax.get_xlim(), ax.get_ylim(), ls="--", c="0.3")
    plt.draw()
    plt.legend(loc=key_loc, shadow=True)
    if outfile is not None:
        plt.savefig(outfile)
    if wait:
        plt.show()


def plot_rocs(evaluations, class_index=0, title=None, key_loc="lower right", outfile=None, wait=True):
    """
    Plots the ROC (receiver operator characteristics) curve for the predictions of multiple classifiers
    on the same dataset.

    TODO: click events http://matplotlib.org/examples/event_handling/data_browser.html

    :param evaluations: the dictionary of Evaluation objects to obtain the predictions from, the key is used
                        in the plot key as prefix
    :type evaluations: dict
    :param class_index: the 0-based index of the class-label to create the plot for
    :type class_index: int
    :param title: an optional title
    :type title: str
    :param key_loc: the position string for the key
    :type key_loc: str
    :param outfile: the output file, ignored if None
    :type outfile: str
    :param wait: whether to wait for the user to close the plot
    :type wait: bool
    """
    if not plot.matplotlib_available:
        logger.error("Matplotlib is not installed, plotting unavailable!")
        return

    if len(evaluations) < 1:
        raise Exception("At least one Evaluation object required!")

    # check if datasets compatible
    if len(evaluations) > 1:
        evaluation = None
        eval_key = None
        for key in evaluations:
            if evaluation is None:
                eval_key = key
                evaluation = evaluations[key]
            else:
                msg = evaluation.header.equal_headers(evaluations[key].header)
                if msg is not None:
                    raise Exception("Evaluation '%s' and '%s' are not compatible: %s" % (eval_key, key, msg))

    ax = None
    for prefix in evaluations:
        evaluation = evaluations[prefix]
        data = generate_thresholdcurve_data(evaluation, class_index)
        head = evaluation.header
        area = get_auc(data)
        x, y = get_thresholdcurve_data(data, "False Positive Rate", "True Positive Rate")
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_xlabel("False Positive Rate")
            ax.set_ylabel("True Positive Rate")
            if title is None:
                title = "ROC"
            ax.set_title(title)
            ax.grid(True)
            fig.canvas.set_window_title(title)
            plt.xlim([-0.05, 1.05])
            plt.ylim([-0.05, 1.05])
        plot_label = prefix + " " + head.class_attribute.value(class_index) + " (AUC: %0.4f)" % area
        ax.plot(x, y, label=plot_label)
        ax.plot(ax.get_xlim(), ax.get_ylim(), ls="--", c="0.3")
    plt.draw()
    plt.legend(loc=key_loc, shadow=True)
    if outfile is not None:
        plt.savefig(outfile)
    if wait:
        plt.show()


def plot_prc(evaluation, class_index=None, title=None, key_loc="lower center", outfile=None, wait=True):
    """
    Plots the PRC (precision recall) curve for the given predictions.

    TODO: click events http://matplotlib.org/examples/event_handling/data_browser.html

    :param evaluation: the evaluation to obtain the predictions from
    :type evaluation: Evaluation
    :param class_index: the list of 0-based indices of the class-labels to create the plot for
    :type class_index: list
    :param title: an optional title
    :type title: str
    :param key_loc: the location string for the key
    :type key_loc: str
    :param outfile: the output file, ignored if None
    :type outfile: str
    :param wait: whether to wait for the user to close the plot
    :type wait: bool
    """
    if not plot.matplotlib_available:
        logger.error("Matplotlib is not installed, plotting unavailable!")
        return
    if class_index is None:
        class_index = [0]
    ax = None
    for cindex in class_index:
        data = generate_thresholdcurve_data(evaluation, cindex)
        head = evaluation.header
        area = get_prc(data)
        x, y = get_thresholdcurve_data(data, "Recall", "Precision")
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_xlabel("Recall")
            ax.set_ylabel("Precision")
            if title is None:
                title = "PRC"
            ax.set_title(title)
            fig.canvas.set_window_title(title)
            plt.xlim([-0.05, 1.05])
            plt.ylim([-0.05, 1.05])
            ax.grid(True)
        plot_label = head.class_attribute.value(cindex) + " (PRC: %0.4f)" % area
        ax.plot(x, y, label=plot_label)
        ax.plot(ax.get_xlim(), ax.get_ylim(), ls="--", c="0.3")
    plt.draw()
    plt.legend(loc=key_loc, shadow=True)
    if outfile is not None:
        plt.savefig(outfile)
    if wait:
        plt.show()


def plot_prcs(evaluations, class_index=0, title=None, key_loc="lower center", outfile=None, wait=True):
    """
    Plots the PRC (precision recall) curve for the predictions of multiple classifiers
    on the same dataset.

    TODO: click events http://matplotlib.org/examples/event_handling/data_browser.html

    :param evaluations: the dictionary of Evaluation objects to obtain the predictions from, the key is used
                        in the plot key as prefix
    :type evaluations: dict
    :param class_index: the 0-based index of the class-label to create the plot for
    :type class_index: int
    :param title: an optional title
    :type title: str
    :param key_loc: the location string for the key
    :type key_loc: str
    :param outfile: the output file, ignored if None
    :type outfile: str
    :param wait: whether to wait for the user to close the plot
    :type wait: bool
    """
    if not plot.matplotlib_available:
        logger.error("Matplotlib is not installed, plotting unavailable!")
        return

    if len(evaluations) < 1:
        raise Exception("At least one Evaluation object required!")

    # check if datasets compatible
    if len(evaluations) > 1:
        evaluation = None
        eval_key = None
        for key in evaluations:
            if evaluation is None:
                eval_key = key
                evaluation = evaluations[key]
            else:
                msg = evaluation.header.equal_headers(evaluations[key].header)
                if msg is not None:
                    raise Exception("Evaluation '%s' and '%s' are not compatible: %s" % (eval_key, key, msg))

    ax = None
    for prefix in evaluations:
        evaluation = evaluations[prefix]
        data = generate_thresholdcurve_data(evaluation, class_index)
        head = evaluation.header
        area = get_prc(data)
        x, y = get_thresholdcurve_data(data, "Recall", "Precision")
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_xlabel("Recall")
            ax.set_ylabel("Precision")
            if title is None:
                title = "PRC"
            ax.set_title(title)
            fig.canvas.set_window_title(title)
            plt.xlim([-0.05, 1.05])
            plt.ylim([-0.05, 1.05])
            ax.grid(True)
        plot_label = prefix + " " + head.class_attribute.value(class_index) + " (PRC: %0.4f)" % area
        ax.plot(x, y, label=plot_label)
        ax.plot(ax.get_xlim(), ax.get_ylim(), ls="--", c="0.3")
    plt.draw()
    plt.legend(loc=key_loc, shadow=True)
    if outfile is not None:
        plt.savefig(outfile)
    if wait:
        plt.show()


def plot_learning_curve(classifiers, train, test=None, increments=100, metric="percent_correct",
                        title="Learning curve", label_template="[#] @ $", key_loc="lower right",
                        outfile=None, wait=True):
    """
    Plots a learning curve.

    :param classifiers: list of Classifier template objects
    :type classifiers: list of Classifier
    :param train: dataset to use for the building the classifier, used for evaluating it test set None
    :type train: Instances
    :param test: optional dataset (or list of datasets) to use for the testing the built classifiers
    :type test: list or Instances
    :param increments: the increments (>= 1: # of instances, <1: percentage of dataset)
    :type increments: float
    :param metric: the name of the numeric metric to plot (Evaluation.<metric>)
    :type metric: str
    :param title: the title for the plot
    :type title: str
    :param label_template: the template for the label in the plot
                           (#: 1-based index of classifier, @: full classname, !: simple classname,
                           $: options, *: 1-based index of test set)
    :type label_template: str
    :param key_loc: the location string for the key
    :type key_loc: str
    :param outfile: the output file, ignored if None
    :type outfile: str
    :param wait: whether to wait for the user to close the plot
    :type wait: bool
    """

    if not plot.matplotlib_available:
        logger.error("Matplotlib is not installed, plotting unavailable!")
        return
    if not train.has_class():
        logger.error("Training set has no class attribute set!")
        return

    if increments >= 1:
        inc = increments
    else:
        inc = round(train.num_instances * increments)

    if test is None:
        tst = [train]
    elif isinstance(test, list):
        tst = test
    elif isinstance(test, Instances):
        tst = [test]
    else:
        logger.error("Expected list or Instances object, instead: " + type(test))
        return
    for t in tst:
        if train.equal_headers(t) is not None:
            logger.error("Training and test set are not compatible: " + train.equal_headers(t))
            return

    steps = []
    cls = []
    evls = {}
    for classifier in classifiers:
        cl = Classifier.make_copy(classifier)
        cls.append(cl)
        evls[cl] = {}
        for t in tst:
            evls[cl][t] = []

    for i in range(train.num_instances):
        if (i > 0) and (i % inc == 0):
            steps.append(i+1)
        for cl in cls:
            # train
            if cl.is_updateable:
                if i == 0:
                    tr = Instances.copy_instances(train, 0, 1)
                    cl.build_classifier(tr)
                else:
                    cl.update_classifier(train.get_instance(i))
            else:
                if (i > 0) and (i % inc == 0):
                    tr = Instances.copy_instances(train, 0, i + 1)
                    cl.build_classifier(tr)
            # evaluate
            if (i > 0) and (i % inc == 0):
                for t in tst:
                    evl = Evaluation(t)
                    evl.test_model(cl, t)
                    evls[cl][t].append(getattr(evl, metric))

    fig, ax = plt.subplots()
    ax.set_xlabel("# of instances")
    ax.set_ylabel(metric)
    ax.set_title(title)
    fig.canvas.set_window_title(title)
    ax.grid(True)
    i = 0
    for cl in cls:
        evlpertest = evls[cl]
        i += 1
        n = 0
        for t in tst:
            evl = evlpertest[t]
            n += 1
            plot_label = label_template.\
                replace("#", str(i)).\
                replace("*", str(n)).\
                replace("@", cl.classname).\
                replace("!", cl.classname[cl.classname.rfind(".") + 1:]).\
                replace("$", join_options(cl.config))
            ax.plot(steps, evl, label=plot_label)
    plt.draw()
    plt.legend(loc=key_loc, shadow=True)
    if outfile is not None:
        plt.savefig(outfile)
    if wait:
        plt.show()
